/* --------------------------------------------------------------------------
 *	Cyrix MediaGX I/O functions
 * --------------------------------------------------------------------------
 *
 *	Copyright (C) 1997		Emmanuel Marty
 *	Copyright (C) 1997		Cyrix Corporation
 *	Copyright (C) 1997		The GGI project
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; see the file COPYING.  If not, write to
 *	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * --------------------------------------------------------------------------
 *
 *	This code was derived from the following sources of information :
 *
 *	Cyrix MediaGX confidential databooks
 *	Cyrix MediaGX gxvideo.c example sourcecode
 *	Help and code from Sengan Baring-Gould of Cyrix Corporation
 *
 *	Many thanks to the Cyrix Corporation for providing me with
 *	documentation and official support about the GX chipset family.
 *
 * --------------------------------------------------------------------------
 *
 *	$Log: gx_mmio.inc,v $
 *	Revision 1.2  1997/12/24 18:21:40  emarty
 *	small fixes and textmode emulation support
 *	
 *	Revision 1.1  1997/12/20 13:02:42  emarty
 *	major: initial checkin
 *	
 */

#define READ_GX32(base,offset) *((uint32 *)((uint8 *)((base)+(offset))))
#define WRITE_GX32(base,offset,value) \
	*((uint32 *)((uint8 *)((base)+(offset)))) = (value);

/*
**	write and read to clockchip registers
*/

void CLK_out8 (struct kgi_display *dpy, uint8 val, uint8 reg) {
	struct	chipset_private *private = KGIM_PRIVATE (dpy, chipset);
	mem_vaddr	gxbase = private->gx_control.baseptr;
	uint32	lockval, gcfg;

	ASSERT (reg <= 0x0E);

	/*	unlock the display controller registers		*/

	lockval = READ_GX32 (gxbase, DC_UNLOCK);
	WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

	/*
	** Before changing the RS lines, the RAMDAC data write buffer
	** must be flushed. This is done by reading from the
	** MC RAMDAC interface
	*/

	gcfg = READ_GX32 (gxbase, MC_RAMDAC_ACC);

	/*	select pll write index register */

	gcfg = READ_GX32 (gxbase, DC_GENERAL_CFG);
	gcfg = (gcfg & ~DC_GCFG_DAC_RS_MASK) |
		( DAC_PLLWRITE << DC_GCFG_DAC_RS_POS );
	WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

	/*	set pll write index */

	WRITE_GX32 (gxbase, MC_RAMDAC_ACC, reg);

	/*	flush RAMDAC write buffer again			*/

	gcfg = READ_GX32 (gxbase, MC_RAMDAC_ACC);

	/*	select pll data register */

	gcfg = READ_GX32 (gxbase, DC_GENERAL_CFG);
	gcfg = (gcfg & ~DC_GCFG_DAC_RS_MASK) |
		( DAC_PLLDATA << DC_GCFG_DAC_RS_POS );
	WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

	/*	set pll index data */

	WRITE_GX32 (gxbase, MC_RAMDAC_ACC, val);

	/*	restore status of lock register			*/

	WRITE_GX32 (gxbase, DC_UNLOCK, lockval);
}

/*
**	this is to prevent compiler optimization from removing
**	our first write since they both go to the same address
*/

void write_ramdac_reg (mem_vaddr gxbase, uint32 value) {
	WRITE_GX32 (gxbase, MC_RAMDAC_ACC, value);     
}

void CLK_out16 (struct kgi_display *dpy, uint16 val, uint8 reg) {
	struct	chipset_private *private = KGIM_PRIVATE (dpy, chipset);
	mem_vaddr	gxbase = private->gx_control.baseptr;
	uint32	lockval, gcfg;

	ASSERT (reg < 0x0E);

	/*	unlock the display controller registers		*/

	lockval = READ_GX32 (gxbase, DC_UNLOCK);
	WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

	/*
	** Before changing the RS lines, the RAMDAC data write buffer
	** must be flushed. This is done by reading from the
	** MC RAMDAC interface
	*/

	gcfg = READ_GX32 (gxbase, MC_RAMDAC_ACC);

	/*	select pll write index register			*/

	gcfg = READ_GX32 (gxbase, DC_GENERAL_CFG);
	gcfg = (gcfg & ~DC_GCFG_DAC_RS_MASK) |
		( DAC_PLLWRITE << DC_GCFG_DAC_RS_POS );
	WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

	/*	set pll write index				*/

	WRITE_GX32 (gxbase, MC_RAMDAC_ACC, reg);

	/*	flush RAMDAC write buffer again			*/

	gcfg = READ_GX32 (gxbase, MC_RAMDAC_ACC);

	/*	select pll data register			*/

	gcfg = READ_GX32 (gxbase, DC_GENERAL_CFG);
	gcfg = (gcfg & ~DC_GCFG_DAC_RS_MASK) |
		( DAC_PLLDATA << DC_GCFG_DAC_RS_POS );
	WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

	/*	set pll index data				*/

	write_ramdac_reg (gxbase, val & 0xFF);
	write_ramdac_reg (gxbase, (val >> 8) & 0xFF);

	/*	restore status of lock register			*/

	WRITE_GX32 (gxbase, DC_UNLOCK, lockval);
}

uint8 CLK_in8 (struct kgi_display *dpy, uint8 reg) {
	struct	chipset_private *private = KGIM_PRIVATE (dpy, chipset);
	mem_vaddr	gxbase = private->gx_control.baseptr;
	uint32	lockval, val, gcfg;

	ASSERT (reg <= 0x0E);

	/*	unlock the display controller registers		*/

	lockval = READ_GX32 (gxbase, DC_UNLOCK);
	WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

	/*
	** Before changing the RS lines, the RAMDAC data write buffer
	** must be flushed. This is done by reading from the
	** MC RAMDAC interface
	*/

	gcfg = READ_GX32 (gxbase, MC_RAMDAC_ACC);

	/*	select pll read index register			 */

	gcfg = READ_GX32 (gxbase, DC_GENERAL_CFG);
	gcfg = (gcfg & ~DC_GCFG_DAC_RS_MASK) |
		( DAC_PLLREAD << DC_GCFG_DAC_RS_POS );
	WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

	/*	set pll read index */

	WRITE_GX32 (gxbase, MC_RAMDAC_ACC, reg);

	/*	flush RAMDAC write buffer again			*/

	gcfg = READ_GX32 (gxbase, MC_RAMDAC_ACC);

	/*	select pll data register */

	gcfg = READ_GX32 (gxbase, DC_GENERAL_CFG);
	gcfg = (gcfg & ~DC_GCFG_DAC_RS_MASK) |
		( DAC_PLLDATA << DC_GCFG_DAC_RS_POS );
	WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

	/*	read pll index value */

	val = READ_GX32 (gxbase, MC_RAMDAC_ACC);

	/*	restore status of lock register			*/

	WRITE_GX32 (gxbase, DC_UNLOCK, lockval);

	return val;
}

/*
**	write and read to RAMDAC registers
*/

void DAC_out8 (struct kgi_display *dpy, uint8 val, uint8 reg) {
	struct	chipset_private *private = KGIM_PRIVATE (dpy, chipset);
	mem_vaddr	gxbase = private->gx_control.baseptr;
	uint32	lockval, gcfg;

	ASSERT (!(reg & 0xF8));

	/*	unlock the display controller registers		*/

	lockval = READ_GX32 (gxbase, DC_UNLOCK);
	WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

	/*	flush RAMDAC write buffer			*/

	gcfg = READ_GX32 (gxbase, MC_RAMDAC_ACC);

	/*	set the register select lines			*/

	gcfg = READ_GX32 (gxbase, DC_GENERAL_CFG);
	gcfg = (gcfg & ~DC_GCFG_DAC_RS_MASK) |
		( (reg & 0x7) << DC_GCFG_DAC_RS_POS );
	WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

	/*	
	**	16bpp mode on a 8-bit bus is possible on MediaGX
	**	if less than 1024 pixels horizontally :
	**	alter RAMDAC mode when conditions are met
	*/

	if ( ((reg & 3) == DAC_PIXEL_MASK) && (val == DAC_MODE_6) &&
	  (dpy->mode.request.virt.x < 1024) )
		val = 0xC0;

	/*	write value to ramdac register			*/

	WRITE_GX32 (gxbase, MC_RAMDAC_ACC, val);
	
	/*	restore status of lock register			*/

	WRITE_GX32 (gxbase, DC_UNLOCK, lockval);
}

uint8 DAC_in8 (struct kgi_display *dpy, uint8 reg) {
	struct	chipset_private *private = KGIM_PRIVATE (dpy, chipset);
	mem_vaddr	gxbase = private->gx_control.baseptr;
	uint32	lockval, gcfg, val;

	ASSERT (!(reg & 0xF8));

	/*	unlock the display controller registers		*/

	lockval = READ_GX32 (gxbase, DC_UNLOCK);
	WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

	/*	flush RAMDAC write buffer			*/

	val = READ_GX32 (gxbase, MC_RAMDAC_ACC);

	/*	set the register select lines			*/

	gcfg = READ_GX32 (gxbase, DC_GENERAL_CFG);
	gcfg = (gcfg & ~DC_GCFG_DAC_RS_MASK) |
	       ( (reg & 0x7) << DC_GCFG_DAC_RS_POS );
	WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

	/*	read value from ramdac register			*/

	val = READ_GX32 (gxbase, MC_RAMDAC_ACC);
	
	/*	restore status of lock register			*/

	WRITE_GX32 (gxbase, DC_UNLOCK, lockval);

	return ((uint8) val);
}

/*
**	set line compare (splitline) value
*/

static void gx_set_split (struct kgi_display *dpy, int line) {
	mem_vaddr gxbase = ((struct chipset_private *)
			   KGIM_PRIVATE (dpy,chipset))->gx_control.baseptr;
	uint32    lockval;

	/*	unlock the display controller registers		*/

	lockval = READ_GX32 (gxbase, DC_UNLOCK);
	WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

	/*	write line compare value			*/
	WRITE_GX32 (gxbase, DC_SS_LINE_COMPARE, line & 0x3FF);

	/*	restore status of lock register			*/

	WRITE_GX32 (gxbase, DC_UNLOCK, lockval);
}

/*
**	set origin
*/

static void gx_set_origin (struct kgi_display *dpy, uint x, uint y) {
	struct	chipset_private *private = KGIM_PRIVATE (dpy, chipset);
	mem_vaddr	gxbase = private->gx_control.baseptr;
	uint32    lockval;

	/*	unlock the display controller registers		*/

	lockval = READ_GX32 (gxbase, DC_UNLOCK);
	WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

	/*	write origin					*/

	WRITE_GX32 (gxbase, DC_FB_ST_OFFSET, 
			y * private->fblinesize +
			x);	/* << pixshift */

	/*	restore status of lock register			*/

	WRITE_GX32 (gxbase, DC_UNLOCK, lockval);
}

/*
**	pointer functions
*/

static void gx_show_pointer(struct kgi_display *dpy, uint x, uint y) {
	struct	chipset_private *private = KGIM_PRIVATE (dpy, chipset);
	mem_vaddr	gxbase = private->gx_control.baseptr;
	uint32	lockval, gcfg;

	/*	unlock the display controller registers		*/

	lockval = READ_GX32 (gxbase, DC_UNLOCK);
	WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

	/*	set cursor position				*/

	WRITE_GX32 (gxbase, DC_CURSOR_X, x & 0x7FF);
	WRITE_GX32 (gxbase, DC_CURSOR_Y, y & 0x3FF);

	/*	enable hardware pointer				*/

	gcfg = READ_GX32 (gxbase, DC_GENERAL_CFG);
	gcfg = gcfg | DC_GCFG_CURE;
	WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

	/*	restore status of lock register			*/

	WRITE_GX32 (gxbase, DC_UNLOCK, lockval);
}

static void gx_hide_pointer(struct kgi_display *dpy) {
	struct	chipset_private *private = KGIM_PRIVATE (dpy, chipset);
	mem_vaddr	gxbase = private->gx_control.baseptr;
	uint32	lockval, gcfg;

	/*	unlock the display controller registers		*/

	lockval = READ_GX32 (gxbase, DC_UNLOCK);
	WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

	/*	disable hardware pointer			*/

	gcfg = READ_GX32 (gxbase, DC_GENERAL_CFG);
	gcfg = gcfg & ~DC_GCFG_CURE;
	WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

	/*	restore status of lock register			*/

	WRITE_GX32 (gxbase, DC_UNLOCK, lockval);
}

#define	gx_undo_pointer	gx_hide_pointer

/*
**	restore display controller registers from provided values
*/

static void restore_state (struct kgi_display *dpy, 
			struct kgi_mode *mode, struct chipsetregs *cr) {
	struct	chipset_private *private = KGIM_PRIVATE (dpy, chipset);
	mem_vaddr	gxbase = private->gx_control.baseptr;
	uint32 lockval, gcfg, tcfg, ocfg, i;

	/*	unlock the display controller registers		*/

	lockval = READ_GX32 (gxbase, DC_UNLOCK);
	WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

	/*	blank display and disable clock			*/

	gcfg = READ_GX32 (gxbase, DC_GENERAL_CFG);
	tcfg = READ_GX32 (gxbase, DC_TIMING_CFG);
	ocfg = READ_GX32 (gxbase, DC_OUTPUT_CFG);

	tcfg &= ~(uint32) DC_TCFG_BLKE;
	WRITE_GX32 (gxbase, DC_TIMING_CFG, tcfg);

	tcfg &= ~(uint32) DC_TCFG_TGEN;
	WRITE_GX32 (gxbase, DC_TIMING_CFG, tcfg);

	/*	disable display fifo load and compression	*/
    
	gcfg &= ~(uint32) (DC_GCFG_DFLE | DC_GCFG_CMPE | DC_GCFG_DECE);
	WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

	/*	clear dotclock multiplication			*/

	gcfg &= ~(uint32) (DC_GCFG_DDCK | DC_GCFG_DPCK | DC_GCFG_DFCK);
	gcfg &= ~(uint32) DC_GCFG_DCLK_MASK;
	WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

	/*	program ramdac and dotclock			*/

	kgim_monitor_set_mode(dpy, mode);
	kgim_clock_set_mode(dpy, mode);
	kgim_ramdac_set_mode(dpy, mode);

	/*	set dotclock multiplication			*/

	gcfg |= cr->general_cfg & DC_GCFG_DCLK_MASK;
	WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

	/*	write display controller regs			*/

	gcfg  = cr->general_cfg;
	gcfg &= ~(uint32) (DC_GCFG_DECE | DC_GCFG_CMPE);
	tcfg  = cr->timing_cfg;
	ocfg  = cr->output_cfg;

	WRITE_GX32 (gxbase, DC_FB_ST_OFFSET, cr->fb_st_offset);
	WRITE_GX32 (gxbase, DC_CB_ST_OFFSET, cr->cb_st_offset);
	WRITE_GX32 (gxbase, DC_CURS_ST_OFFSET, cr->cur_st_offset);
	WRITE_GX32 (gxbase, DC_LINE_DELTA, cr->dc_line_delta);
	WRITE_GX32 (gxbase, DC_BUF_SIZE, cr->dc_buf_size);

	WRITE_GX32 (gxbase, DC_H_TIMING_1, cr->h_timing_1);
	WRITE_GX32 (gxbase, DC_H_TIMING_2, cr->h_timing_2);
	WRITE_GX32 (gxbase, DC_H_TIMING_3, cr->h_timing_3);
	WRITE_GX32 (gxbase, DC_FP_H_TIMING, cr->h_timing_3);
	WRITE_GX32 (gxbase, DC_V_TIMING_1, cr->v_timing_1);
	WRITE_GX32 (gxbase, DC_V_TIMING_2, cr->v_timing_2);
	WRITE_GX32 (gxbase, DC_V_TIMING_3, cr->v_timing_3);
	WRITE_GX32 (gxbase, DC_FP_V_TIMING, cr->v_timing_3);

	WRITE_GX32 (gxbase, DC_OUTPUT_CFG, ocfg);
	WRITE_GX32 (gxbase, DC_TIMING_CFG, tcfg);
	WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);

	/*	prepare and enable compression if requested	*/

	if (cr->general_cfg & (DC_GCFG_DECE|DC_GCFG_CMPE)) {

		/*	clear compressed line "valid" bits	*/

		for (i=0; i < 1024; i++) {
			WRITE_GX32 (gxbase, MC_DR_ADD, i);
			WRITE_GX32 (gxbase, MC_DR_ACC, 0);
		}

		/*	enable framebuffer compression		*/

		gcfg |= (DC_GCFG_DECE|DC_GCFG_CMPE);
		WRITE_GX32 (gxbase, DC_GENERAL_CFG, gcfg);
	}

	/*	write line compare and cursor registers		*/

	WRITE_GX32 (gxbase, DC_SS_LINE_COMPARE, cr->ss_line_compare);
	WRITE_GX32 (gxbase, DC_CURSOR_X, cr->dc_cursor_x);
	WRITE_GX32 (gxbase, DC_CURSOR_Y, cr->dc_cursor_y);
	WRITE_GX32 (gxbase, DC_CURSOR_COLOR, cr->dc_cursor_color);

	/*	restore status of lock register			*/

	WRITE_GX32 (gxbase, DC_UNLOCK, lockval);
}
