/* --------------------------------------------------------------------------
 * IBM (VGA) I/O
 * --------------------------------------------------------------------------
 *
 * Copyright (C) 1995 by The Linux-GGI Project
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ---------------------------------------------------------------------------
 *
 * $Log: vga_io.inc,v $
 * Revision 1.10  1997/12/14 21:55:40  becka
 * uint/sint namespace fixed ... please test ...
 *
 * Revision 1.9  1997/12/09 10:11:38  seeger
 * Global cleanup for new include directory locations.
 *
 * Revision 1.8  1997/09/13 00:12:23  taylorj
 * *** empty log message ***
 *
 * Revision 1.7  1997/09/06 04:05:44  taylorj
 * Major: New maintainer (me).  Also reformatted all VGA driver source to GGI
 * specs and converted all debugging prints to use the newer debugging macros.
 *
 * Revision 1.6  1997/04/17 06:17:10  ezrec
 * Minor: Fixes to work with the new system.h and module.h files...
 *
 * Revision 1.5  1997/04/10 03:34:03  ezrec
 * Major: More fixes, and I'm attempting (ha!) to use Steffen's new MMAP scheme..
 *
 * Revision 1.4  1997/03/05 20:48:43  ezrec
 * Minor: VGA base region no longer conflict with MDA driver...
 *
 * Revision 1.3  1997/02/20 21:59:07  ezrec
 * Minor: jmcc fixes: keypress_wait bug on 2.1.x (console.c)
 * Jochen Hoenicke <Jochen.Hoenicke@Informatik.Uni-Oldenburg.DE> fixed:
 * IBM/vga drivers
 * No-display-on-driver-load bug (woo woo!)
 * set_origin/Attribute handling bugs...
 * Added TEXT_A000_64, for those of you with MDA or
 * Hercules cards (it's now the default, anyway)
 *
 * Revision 1.2  1997/02/04 04:33:45  ezrec
 * Minor: Small bugfixes to cursor and pointer code, but not fixed yet.
 *
 * Revision 1.1  1997/02/03 15:06:35  ezrec
 * Major: IBM vga driver (initial version, taken from GGI 0.0.8)
 *
 * ---------------------------------------------------------------------------
 * This code was derived from the following source of information:
 * [1]	XFree86 vga driver
 */

/* GGI includes */

#include "IBM/vga_io.inc"

/* Generalized DAC output functions */

static unsigned short dac_port[4] =
{ 
	0x3C8, 0x3C9, 0x3C6, 0x3C7 
};

void DAC_out8(struct kgi_display *dpy, uint8 value, uint8 dac_reg)
{
	if (dac_reg & ~0x3) 
	{
		
		DEBUG1("DAC register select lines truncated: %x to %x",
		       dac_reg, dac_reg &= 0x3);
	}
	
	io_out8(value, dac_port[dac_reg & 0x03]);
}

uint8 DAC_in8(struct kgi_display *dpy, uint8 dac_reg)
{
	if(dac_reg & ~0x03) 
	{
		
		DEBUG1("DAC register select lines truncated: %x to %x",
		       dac_reg, dac_reg &= 0x03);
	}
	
	return io_in8(dac_port[dac_reg & 0x03]);
 }

void DAC_ins8(struct kgi_display *dpy,uint8 dac_reg,void *buf, ggi_uint count)
{
	if(dac_reg & ~0x03)
	{
		
		DEBUG1("DAC register select lines truncated: %x to %x",
		       dac_reg, dac_reg &= 0x3);
	}

	io_ins8(dac_port[dac_reg & 0x03], buf, count);
}

void DAC_outs8(struct kgi_display *dpy,uint8 dac_reg, void *buf, ggi_uint count)
{
	if(dac_reg & ~0x03) 
	{
		
		DEBUG1("DAC register select lines truncated: %x to %x",
		       dac_reg, dac_reg &= 0x03);
	}

	io_outs8(dac_port[dac_reg & 0x03], buf, count);
}

/* Clock I/O functions */

void CLK_out8(struct kgi_display *dpy, uint8 value, uint8 reg)
{
	if (reg != 0)
	{
		
		INTERNALERROR;
	}
	
	switch (value) {
	case 0: 
		value = MISC_25MHZ_CLK;
		break;
	case 1:
		value = MISC_28MHZ_CLK;
		break;
	default:
		INTERNALERROR;
		return;
	}

	io_out8((io_in8(MISCr) & 0xf3) | value, MISCw);
}

uint8 CLK_in8(struct kgi_display *dpy,uint8 reg)
{
	if (reg != 0)
	{
		
		INTERNALERROR;
	}
	

	return (io_in8(MISCr) & 0x04) >> 2;
}

#define NoSEQRegs	0x05
#define	NoCRTRegs	0x19
#define NoGRCRegs	0x09
#define NoATCRegs	0x15

struct statetable
{
	uint8 MISC;
	uint8 SEQ[NoSEQRegs];
	uint8 CRT[NoCRTRegs];
	uint8 GRC[NoGRCRegs];
	uint8 ATC[NoATCRegs];
};

static void save_SEQ(uint8 *buf)
{
	int i;

	for (i = 0; i < NoSEQRegs; i++) 
	{
		
		*(buf++) = inbSEQ(i);
	}
}

static void restore_SEQ(uint8 *buf)
{
	register uint8 i;
	for (i = 0; i < NoSEQRegs; i++) 
	{
		
		outbSEQ(*(buf++), i);
	}
}

static void save_CRT(uint8 *buf)
{
	int i;

	for (i = 0; i < NoCRTRegs; i++) 
	{
		
		*(buf++) = inbCRT(i);
	}
}

static void restore_CRT(uint8 *buf)
{
	register uint8 i;

	outbCRT(inbCRT(CRT_VSYNCEND) & ~CR11_LOCKTIMING, CRT_VSYNCEND);
	outbCRT(inbCRT(CRT_HBLANKEND) | CR03_IS_VGA,CRT_HBLANKEND);

	for (i = 0; i < NoCRTRegs; i++) 
	{
		switch (i) 
		{
		case CRT_HBLANKEND:
			*buf |= CR03_IS_VGA;
			break;
		case CRT_VSYNCEND:
			*buf &= ~CR11_LOCKTIMING;
			break;
		default:
			break;
		}
		outbCRT(*(buf++), i);
	}
	
}

static void save_GRC(uint8 *buf)
{
  int i;

	for (i = 0; i < NoGRCRegs; i++) 
	{
		
		*(buf++) = inbGRC(i);
	}
}

static void restore_GRC(uint8 *buf)
{
	register uint8 i;

	for (i = 0; i < NoGRCRegs; i++) 
		outbGRC(*(buf++), i);
}

static void save_ATC(uint8 *buf)
{
	int i;

	for (i = 0; i < NoATCRegs; i++) 
	{
		io_in8(INS1r);  /* Reset AFF and setup index */
		*(buf++) = inbATC(i);
	}
	io_in8(INS1r); 
	io_out8(ATCI_ENABLEDISPLAY, ATCI);
}

static void restore_ATC(uint8 *buf)
{
	register uint8 i;

	for (i = 0; i < NoATCRegs; i++) {
		io_in8(INS1r); 
		outbATC(*(buf++), i);
	}
	io_in8(INS1r); 
	io_out8(ATCI_ENABLEDISPLAY, ATCI);
}

static void save_state(struct statetable *state) 
{
	state->MISC = io_in8(MISCr);	/* Save miscellaneous output reg */
	
	save_SEQ(state->SEQ);	/* Save sequencer regs */
	save_CRT(state->CRT);	/* Save CRT controller regs */
	save_GRC(state->GRC);	/* Save graphics controller regs */
	save_ATC(state->ATC);	/* Save attribute controller regs */
}

static void restore_state(struct statetable *state)
{
	outbMISC((io_in8(MISCr) & MISC_CLOCK_MASK) | state->MISC);

	restore_SEQ(state->SEQ);
	restore_CRT(state->CRT);
	restore_GRC(state->GRC);
	restore_ATC(state->ATC);
}

/* FIXME: Move this over with the other KGI debugging stuff */
#if DEBUG_LEVEL > 0
static void dump_state(struct statetable *state)
{ 
	int i;
	printk("MISC: %.2x", state->MISC);
  
	for (i = 0; i < NoSEQRegs; i++)
	{ 
		if(!(i & 0xF))
			printk("\nSEQ:%.2x: ", i);
		printk("%.2x, ", state->SEQ[i]);
	}
	
	for (i = 0; i < NoCRTRegs; i++)
	{ 
		if(!(i & 0xf))
			printk("\nCRT:%.2x: ", i);
		printk("%.2x, ", state->CRT[i]);
	}
	
	for (i = 0; i < NoGRCRegs; i++)
	{ 
		if(!(i & 0xF)) 
			printk("\nGRC:%.2x: ", i);
		printk("%.2x, ", state->GRC[i]);
	}
	
	for (i = 0; i < NoATCRegs; i++)
	{ 
		if(!(i & 0xF)) 
			printk("\nATC:%.2x: ", i);
		printk("%.2x, ", state->ATC[i]);
	}
	printk("\n");
}
#endif
