#ifndef _mmio_inc_INC_
#define _mmio_inc_INC_

#define VGA_MAXPAGE 16

#include "IBM/vga.h"
#include "IBM/vga_io.inc"
#include "IBM/modes.h"
#include <linux/mm.h>

#define MM_PROT_RW (VM_READ|VM_WRITE|VM_SHARED)
#define MM_PROT_W  (VM_WRITE|VM_SHARED)

/* Set the page to map and clear old mappings */
static void mmap_setpage_4bit(struct kgi_mmio_region *mm, ggi_uint offset)
{
	int page = (offset >> 16) & 0xF;

	DEBUG1("chipset_mmap_setpage: setting 4bit page to %d", page);
	outbGRC(page & 3, 4); /* Read Map select */
	outbGRC(page & 15, 0);/* Color select */

	mm->offset = (page << 16);
}

static void mmap_setpage_modex(struct kgi_mmio_region *mm,ggi_uint offset)
{
	int page = (offset >> 16) & 0xF;

	DEBUG1("chipset_mmap_setpage: setting 8bit page to %d", page);
	outbSEQ(page, SEQ_WPLANE); /* Write Map */
	outbGRC(page & 3, 4); /* Read Map select */

	mm->offset = (page << 16);
}

static struct kgi_mmio_region mm_text16 = 
{
	NULL,			/* Next					*/
	"text16",		/* Name 				*/
	MMAP_FRAMEBUFFER,	/* Type (Linear)			*/
	MM_PROT_RW,		/* Protection map			*/
	0x10000,		/* Total size of area			*/
	0x10000,		/* Physical size (64k)			*/
	0xA0000,		/* Physical address of the windows	*/
	0,			/* Logical offset			*/
	NULL			/* Set offset function			*/
};
	
static struct kgi_mmio_region mm_1bit = 
{
	NULL,			/* Next					*/
	"VGA 1bit",		/* Name					*/
	MMAP_FRAMEBUFFER,	/* Type (Linear)			*/
	MM_PROT_RW,		/* Protection map			*/
	0x10000,		/* Total size of area			*/
	0x10000,		/* Physical size (64k)			*/
	0xA0000,		/* Physical address of the windows	*/
	0,			/* Logical offset			*/
	NULL			/* Set offset function			*/
};

static struct kgi_mmio_region mm_4bit = 
{
	NULL,			/* Next					*/
	"VGA 4bit",		/* Name 				*/
	MMAP_IBM_VGA_4,		/* Type (VGA 4bit planes)		*/
	MM_PROT_RW,		/* Protection map			*/
	16 * 0x10000,		/* Total size of area			*/
	0x10000,		/* Physical size (64k)			*/
	0xA0000,		/* Physical address of the windows	*/
	0,			/* Logical offset			*/
	mmap_setpage_4bit	/* Set offset function			*/
};
	
static struct kgi_mmio_region mm_8bit = 
{
	NULL,			/* Next					*/
	"8bit",			/* Name					*/
	MMAP_FRAMEBUFFER,	/* Type	(Linear)			*/
	MM_PROT_RW,		/* Protection map			*/
	0x10000,		/* Total size of area			*/
	0x10000,		/* Physical size (64k)			*/
	0xA0000,		/* Physical address of the windows	*/
	0,			/* Logical offset			*/
	NULL			/* Set offset function			*/
};

static struct kgi_mmio_region mm_modex = 
{
	NULL,			/* Next					*/
	"modex",		/* Name					*/
	MMAP_IBM_VGA_MODEX,	/* Type (Mode X)			*/
	MM_PROT_RW,		/* Protection map			*/
	16 * 0x10000,		/* Total size of area			*/
	0x10000,		/* Physical size (64k)			*/
	0xA0000,		/* Physical address of the windows	*/
	0,			/* Logical offset			*/
	mmap_setpage_modex	/* Set offset function			*/
};

static int graphics_mmio_setup(struct kgi_display *dpy, struct kgi_mode *mode)
{
	struct kgi_mmio_region *mm = NULL;

	switch (mode->request.graphtype) {
		case GT_TEXT16:
		mm = &mm_text16;
		break;
		case GT_1BIT:
		mm = &mm_1bit;
		break;
		case GT_4BIT:
		mm = &mm_4bit;
		break;
		case GT_8BIT:
		if (mode->request.virt.x == 320 && 
		    mode->request.virt.y == 200)
			mm = &mm_8bit;
		else
			mm = &mm_modex;
		break;
		default:
		return -E(GRAPHICS,BOGUS);
	}

	if (mm != NULL) {
		mm->next = mode->mmio;
		mode->mmio = mm;
	}

	return EOK;
}

#endif /* _mmio_inc_ */
