/* ---------------------------------------------------------------------------
 * S3 generic acceleration driver
 * ---------------------------------------------------------------------------
 *
 * Copyright (C) 1995-97, Andreas Beck and Steffen Seeger
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * ---------------------------------------------------------------------------
 * 
 * $Log: generic.c,v $
 * Revision 1.13  1997/12/14 21:56:06  becka
 * uint/sint namespace fixed ... please test ...
 *
 * Revision 1.12  1997/11/28 02:17:59  becka
 * GC mapping working.
 *
 * Revision 1.11  1997/11/08 15:12:46  becka
 * Namespace cleanup ... probably broke all and everything ...
 *
 * Revision 1.10  1997/09/12 07:28:45  taylorj
 * Minor: Reformatted source to be consistent with GGI coding style guidelines,
 * modified all debugging output to use the DEBUGx macros, and added CVS
 * revision logging to headers where needed.
 *
 * 
 * ---------------------------------------------------------------------------
 */

#include <ggi/maintainers.h>
#define	MAINTAINER	Andreas_Beck
#define	DRIVER_REV	"$Revision: 1.13 $"

#define DEBUG_LEVEL	0

#undef	DEBUG_IO

/*\
|*| Kernel includes 
\*/

#include <asm/io.h>
#include <linux/ioport.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#define	request_io_region	request_region
#define	check_io_region		check_region

/*\
|*| KGI includes
\*/

#include <ggi/module.h>
#include <ggi/ggi_commands.h>
#include <ggi/graphic.h>

#include "suggest.inc"

#define	VISION964
#define VISION968
#include "S3/Vision.h"

#include "S3/accel.h"

#define NUMPINGPONGPAGES 2

void *pingpong;
enum pingpongstatus 
{ 
	WAITING, RECEIVING, EXECUTING 
};

enum pingpongstatus ppstat[NUMPINGPONGPAGES];

/* Helper functions */
int s3_wait_fifo(int spc_need, int maysched)
{	
	int x = 1000000000;
	while((inb(ECR_GP_STATUS) & (0x100 >> spc_need)) && --x) 
		if (maysched) schedule();
	return(!x);
}

static int s3_wait_idle(int maysched)
{ 	
	int x = 1000000000;
	while((inw(ECR_GP_STATUS) & EC06_GE_BUSY) && --x) 
		if (maysched) schedule();
	return(!x);
}

/* Accelerated drawing functions follow */
static int s3_drawbox(struct kgi_graphic *graphic, void *iobuf, int maysched)
{	
	struct box *boxp = (struct box *)iobuf;

	if (graphic->dev.mode.graphtype == GT_4BIT || 
	    graphic->dev.mode.graphtype == GT_TEXT16)
	{
		
		return -E(GRAPHICS,NOSUP_MODE_MMAP);
	}

	if (boxp->w && boxp->h) {	/* 0 width is not OK! */

		if (s3_wait_fifo(3, 1))
		{
			
			return -E(GRAPHICS,NOSUP_NOW_MMAP);
		}
		
		outw(0x27, ECR_FOREGR_MIX);
		outl(graphic->gc->fg_color, ECR_FOREGROUND);
		outw(MF_PIX_CONTROL, ECR_MULTIFUNC);

		if (s3_wait_fifo(5, maysched))
		{
			
			return -E(GRAPHICS,NOSUP_NOW_MMAP);
		}
		
		outw(boxp->x, ECR_CUR_X);
		outw(boxp->y, ECR_CUR_Y);
		outw(boxp->w - 1, ECR_MAJAXIS_CNT);
		outw((boxp->h - 1) | MF_MINAXIS_CNT, ECR_MULTIFUNC);
		outw(EC07_FILLRECT | EC07_LITTLE_ENDIAN |  EC07_32BIT_IO |
		     EC07_DIR(5) | EC07_DRAWPIXEL | EC07_ACROSS_PLANE |
	             0x0001, ECR_COMMAND);
		s3_wait_idle(maysched);
	}

	return(0);
}

static int s3_copybox(struct kgi_graphic *graphic,void *iobuf)
{	
	struct cbox *boxp = (struct cbox *)iobuf;
	int dir = 0;

	if (graphic->dev.mode.graphtype == GT_4BIT || 
	    graphic->dev.mode.graphtype == GT_TEXT16)
		return -E(GRAPHICS,NOSUP_MODE_MMAP);

	if (boxp->w&&boxp->h) {		/* 0 width is not OK! */

		if (s3_wait_fifo(2, 1)) 
		{
			
			return -E(GRAPHICS,NOSUP_NOW_MMAP);
		}

		outw(0x67, ECR_FOREGR_MIX);
		outw(MF_PIX_CONTROL, ECR_MULTIFUNC);

		if (boxp->x > boxp->x2) 
		{
			
			dir |= EC07_DIR(1);
		}
		else 
		{
			
			boxp->x += boxp->w - 1;
			boxp->x2 += boxp->w - 1;
		}

		if (boxp->y > boxp->y2) 
		{
			
			dir |= EC07_DIR(4);
		}
		else 
		{ 
			boxp->y += boxp->h - 1;
			boxp->y2 += boxp->h - 1;
		}

		if (s3_wait_fifo(7, 1)) 
		{
			
			return -E(GRAPHICS,NOSUP_NOW_MMAP);
		}
		
		outw(boxp->x, ECR_CUR_X);
		outw(boxp->y, ECR_CUR_Y);
		outw(boxp->x2, ECR_DIAG_STEP);
		outw(boxp->y2, ECR_AXIAL_STEP);
		outw(boxp->w - 1, ECR_MAJAXIS_CNT);
		outw((boxp->h - 1) | MF_MINAXIS_CNT,ECR_MULTIFUNC);
		outw(EC07_BITBLT | EC07_LITTLE_ENDIAN | EC07_32BIT_IO |
		     dir | EC07_DRAWPIXEL | EC07_ACROSS_PLANE |
	             0x0001, ECR_COMMAND);
		s3_wait_idle(1);
	}
	
	return(0);
}

static int s3_drawline(struct kgi_graphic *graphic, void *iobuf)
{ 
	int dx, dy, min, max;
	struct line *linp = (struct line *)iobuf;

	if (graphic->dev.mode.graphtype == GT_4BIT || 
	    graphic->dev.mode.graphtype == GT_TEXT16)
	{
		
		return -E(GRAPHICS,NOSUP_MODE_MMAP);
	}

	dx = linp->x2 - linp->x;
	dy = linp->y2 - linp->y;

	if (abs(dx) > abs(dy))
	{
		
		max = abs(dx);
		min = abs(dy);
	}
	else
	{
		
		max = abs(dy);
		min = abs(dx);
	}

	if (max == 0)
	{
		
		return(0);	/* FIXME! */
	}

	if (s3_wait_fifo(3, 1))
	{ 
		
		return -E(GRAPHICS,NOSUP_NOW_MMAP);
	}

	outw(0x27, ECR_FOREGR_MIX);
	outl(graphic->gc->fg_color, ECR_FOREGROUND);
	outw(MF_PIX_CONTROL, ECR_MULTIFUNC);

	if (s3_wait_fifo(7, 1))
	{
		
		return -E(GRAPHICS,NOSUP_NOW_MMAP);
	}
	
	outw(linp->x, ECR_CUR_X);
	outw(linp->y, ECR_CUR_Y);

	outw((max - 1), ECR_MAJAXIS_CNT);
	outw((min - max) << 1, ECR_DIAG_STEP);
	outw((min << 1), ECR_AXIAL_STEP);

	outw((min + min - max - (dx < 0)), ECR_ERROR_TERM);

	outw(EC07_LINE
	        | ((dy>0)            ? EC07_DIR(4) : 0)
		| ((dx>0)            ? EC07_DIR(1) : 0)
		| ((abs(dy)>abs(dx)) ? EC07_DIR(2) : 0)
		| EC07_DRAWPIXEL 
		| EC07_ACROSS_PLANE
		| 0x0001 ,ECR_COMMAND);

	s3_wait_idle(1);

	return(0);
}

/* KGI module interface */

int kgim_graphics_init(struct kgi_card *card, struct kgi_display *dpy)
{
	NOTICE0("S3 accel driver loaded");
	/* !!! Check for Chipset and Version here !!! */

	return EOK;
}

void kgim_graphics_done(struct kgi_display *dpy)
{
	NOTICE1("S3 accel driver removed");
}

int kgim_graphics_check_mode (struct kgi_display *dpy, struct kgi_mode *mode,
			      enum kgi_tm_cmd cmd)
{
	/* !!! We should probably check for accel-friendly dimensions here */
	return EOK;
}

void kgim_graphics_set_mode (struct kgi_display *dpy, struct kgi_mode *mode)
{

	if (mode->request.graphtype != GT_TEXT16) {

		/* Reset the Accelerator Engine */
		outw(((mode->request.graphtype == GT_4BIT) 
		      ? (EC01_4BPP|EC01_ENH_ENB) : EC01_ENH_ENB) | EC01_LAW_ENB, ECR_ADVFUNC_CNTL);

		outw(EC00_RESET, ECR_SUBSYS_CNTL);
		inw_p(ECR_SUBSYS_STAT);	/* Wait a bit */
		outw(EC00_ENABLE, ECR_SUBSYS_CNTL);
		inw_p(ECR_SUBSYS_STAT); /* Wait a bit */

		outw(MF_MULT_MISC, ECR_MULTIFUNC); /* 16 bit I/O registers */

		/* Now set some basic accelerator registers */
		/* This might run in interrupt context, so maysched is 0 */
		s3_wait_fifo(6, 0);
		outw(SRC_FOREGROUND | MIX_SRC, ECR_FOREGR_MIX);
		outw(SRC_BACKGROUND | MIX_SRC, ECR_BACKGR_MIX);/* Direct color*/

		outw(MF_T_CLIP | 0, ECR_MULTIFUNC);	/* Clip virtual area  */
		outw(MF_L_CLIP | 0, ECR_MULTIFUNC);
		outw(MF_R_CLIP | (mode->request.virt.x - 1), ECR_MULTIFUNC);
		outw(MF_B_CLIP | (mode->request.virt.y - 1), ECR_MULTIFUNC);

		s3_wait_fifo(5, 0);
		outw(~0, ECR_WRITE_MASK);	/* Enable all planes */
		outw(~0, ECR_WRITE_MASK);	/* Enable all planes */
		outw(MF_PIX_CONTROL | MFA_SRC_FOREGR_MIX, ECR_MULTIFUNC);

		{
			struct box boxp;
			boxp.x = boxp.y = 0;
			boxp.w = mode->request.virt.x;
			boxp.h = mode->request.virt.y;
			s3_drawbox((struct kgi_graphic *) &dpy->dev, &boxp, 0);
		}
		/* !!! reset other registers too. */
	}
	return;
}

int kgim_graphics_command (struct kgi_graphic *gr, ggi_uint cmd, ggi_uint arg)
{ 
	ggi_suggest *sugg;
	const char *name;

	if (!(gr->dev.state & KGI_MAPPED_ON_DISPLAY) ) {

		return -E(GRAPHICS,NOSUP_MODE_MMAP);
		/* Don't accel inactive screen! */
	}

	switch (cmd) { 
		
	case GRAPHICS_GETSUGGEST:
		sugg = (ggi_suggest *)(gr->io_buf);
		gr->io_res = gr->io_buf;

		if (sugg->id < 0 || sugg->id > 4) 
		{
			
			return -E(GRAPHICS,BOGUS);
		}
		
		name = sugg_default[gr->dev.mode.graphtype].text[sugg->id];
		if (!name)
		{
			
			return -E(GRAPHICS,BOGUS);
		}
		
		strcpy(sugg->name, name);
		sugg->args[0] = 0;
		sugg->id++;	/* Next is RAMDAC */
		if (sugg->id >= 4 || sugg_default[gr->dev.mode.graphtype].text[sugg->id] == NULL) 
		{
			
			sugg->id = 0;
		}
		
		return 0;

	case GRAPHICS_DRAWPIXEL:
#if 0	/* 1=SLOW */
		{  
			struct ggi_coord *h = gr->io_buf;
			unsigned offset = h->x + h->y * current_mode.virt.x;
			int col = 0x12345678;

			switch(current_mode.graphtype) {
			case GT_8BIT:
			 	((char *)fb_graphics)[offset] = (char)col;
				return EOK;
			
			case GT_15BIT:
			case GT_16BIT:	
				((short *)fb_graphics)[offset] = (short)col;
				return EOK;

			case GT_32BIT:
				((int *)fb_graphics)[offset] = (int)col;
				return EOK;
			default:
				printk("DUNNO MYSELF ... ???");
			}
		} 
		break;
#else
		return -E(GRAPHICS,NOSUP_ALWAYS_MMAP);
#endif

		/* Really accelerated functions. */

	case GRAPHICS_DRAWBOX:
		return(s3_drawbox(gr, gr->io_buf, 1));
		
	case GRAPHICS_COPYBOX:
		return(s3_copybox(gr, gr->io_buf));
		
	case GRAPHICS_DRAWLINE:
		return(s3_drawline(gr, gr->io_buf));

		/* Functions that will probably benefit 
		** from lower-lever accel 
		*/
		
	case GRAPHICS_FILLSCREEN:
		return -E(GRAPHICS,NOSUP_ALWAYS_LOWER);

		/* Functions that should probably use MMAP
		 */
	default: 
		return -E(GRAPHICS,NOSUP_ALWAYS_MMAP);

/* These functions cannot be supported via standard interfaces due to
 * pointers to userspace in the args, but are efficiently emulated via MMAP:
 * GRAPHICS_PUTHLINE	GRAPHICS_GETHLINE	GRAPHICS_PUTVLINE
 * GRAPHICS_GETVLINE	GRAPHICS_GETBOX		GRAPHICS_PUTBOX
 */

/* These are not implemented due to efficient support via MMAP.
 * GRAPHICS_PUTPIXEL	GRAPHICS_GETPIXEL	GRAPHICS_DRAWPIXEL (dummy for testing above)
 * GRAPHICS_DRAWCIRCLE	GRAPHICS_DRAWHLINE	GRAPHICS_DRAWVLINE
 */
	}
	return -E(GRAPHICS,NOSUP_ALWAYS_CANT);	/* The driver can't do this */
}
