/* --------------------------------------------------------------------------
**	IBM RGB526(DB) ramdac register definitions
** --------------------------------------------------------------------------
**
**	Copyright (C) 1997 Torgeir Veimo
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
** 
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
** 
** ---------------------------------------------------------------------------
** 
**	This code was derived from the following source of information:
**
**	[1]	XFree86 X server source code
**	[2]	IBM RGB526(DB) Databook (Preliminary).
**	[3]	IBM RGB 524-526-624 Palette DAC Conversion - Application Note
**
**	MAINTAINER	Torgeir_Veimo
*/
#ifndef _IBM_RGB526_H
#define _IBM_RGB526_H

#include "IBM/vga_dac.h"

/*
**	Extended registers
*/
#define DAC_INDEX_LOW	0x04
#define DAC_INDEX_HIGH	0x05
#define DAC_INDEX_DATA	0x06
#define DAC_INDEX_CTRL	0x07

/*
**	Internal registers. Accessed through the extended registers
*/

/*	Revision Level, preferred values.
**
**	The RGB524 has two different revision levels:
**
**	IBM37RGB524CF 17/22-A: SCLK INVT operate correctly: rev. level 0xE0 
**	IBM37RGB524CF 17/22: SCLK INVT doesn't operate correctly: rev.lvl 0xF0
**
**	The other ramdacs SCLK INVT operates correctly.
*/

#define DAC_REV			0x00
	#define _RGB524_REV_LEVEL	0xE0
	#define _RGB524_REV_BAD		0xF0 
	#define _RGB525_REV_LEVEL	0x80
	#define _RGB526_REV_LEVEL	0xC0
	#define _RGB624_REV_LEVEL	0xC0

/*	Product ID
**
**	The RGB624 has a different product id to identify that YUV pixel 
**	formats are available.
*/

#define DAC_ID			0x01
	#define _RGB524_PROD_ID		0x02
	#define _RGB525_PROD_ID		0x02
	#define _RGB526_PROD_ID		0x02
	#define _RGB624_PROD_ID		0x30

#define DAC_MISC_CLOCK    	0x02
	#define _RGB52X_DDOTCLK_ENABLE		(0 << 7)
	#define _RGB52X_DDOTCLK_DISABLE		(1 << 7)
	#define _RGB52X_SCLK_ENABLE		(0 << 6)
	#define _RGB52X_SCLK_DISABLE		(1 << 6)
	#define _RGB52X_B24P_DDOT_DIV_PLL	(0 << 5)
	#define _RGB52X_B24P_DDOT_SCLK		(1 << 5)
	#define _RGB52X_DDOT_PLL_DIV_1		(0 << 1)
	#define _RGB52X_DDOT_PLL_DIV_2		(1 << 1)
	#define _RGB52X_DDOT_PLL_DIV_4		(2 << 1)
	#define _RGB52X_DDOT_PLL_DIV_8		(3 << 1)
	#define _RGB52X_DDOT_PLL_DIV_16		(4 << 1)
	#define _RGB52X_PLL_DISABLE		(0 << 0)
	#define _RGB52X_PLL_ENABLE		(1 << 0)

/*
**	Sync Control
*/
#define DAC_SYNC		0x03
	#define _RGB52X_DLY_CTRL_ADD		(0 << 7)
	#define _RGB52X_DLY_SYNC_NOADD		(1 << 7)
	#define _RGB52X_CSYN_INVT_DISABLE	(0 << 6)
	#define _RGB52X_CSYN_INVT_ENABLE	(1 << 6)
	#define _RGB52X_VSYN_INVT_DISABLE	(0 << 5)
	#define _RGB52X_VSYN_INVT_ENABLE	(1 << 5)
	#define _RGB52X_HSYN_INVT_DISABLE	(0 << 4)
	#define _RGB52X_HSYN_INVT_ENABLE	(1 << 4)
	#define _RGB52X_VSYN_CTRL_NORMAL	(0 << 2)
	#define _RGB52X_VSYN_CTRL_HIGH		(1 << 2)
	#define _RGB52X_VSYN_CTRL_LOW		(2 << 2)
	#define _RGB52X_VSYN_VTRL_DISABLE	(3 << 2)
	#define _RGB52X_HSYN_CTRL_NORMAL	(0 << 0)
	#define _RGB52X_HSYN_CTRL_HIGH		(1 << 0)
	#define _RGB52X_HSYN_CTRL_LOW		(2 << 0)
	#define _RGB52X_HSYN_CTRL_DISABLE	(3 << 0)

/*
**	HSync Control
*/
#define DAC_HSYNC_POS		0x04
	#define _RGB52X_HSYN_POS(n)		((n) & 0xF)

/*
**	Power Managment
*/
#define DAC_PWR_MGMT		0x05
	#define _RGB52X_SCLK_PWR_NORMAL		(0 << 4)
	#define _RGB52X_SCLK_PWR_DISABLE	(1 << 4)
	#define _RGB52X_DDOT_PWR_NORMAL		(0 << 3)
	#define _RGB52X_DDOT_PWR_DISABLE	(1 << 3)
	#define _RGB52X_SYNC_PWR_NORMAL		(0 << 2)
	#define _RGB52X_SYNC_PWR_DISABLE	(1 << 2)
	#define _RGB52X_ICLK_PWR_NORMAL		(0 << 1)
	#define _RGB52X_ICLK_PWR_DISABLE	(1 << 1)
	#define _RGB52X_DAC_PWR_NORMAL		(0 << 0)
	#define _RGB52X_DAC_PWR_DISABLE		(1 << 0)

/*
**	DAC Operation
*/
#define DAC_DAC_OP		0x06
	#define _RGB52X_SOG_DISABLE		(0 << 3)
	#define _RGB52X_SOG_ENABLE		(1 << 3)
	#define _RGB52X_BRB_DISABLE		(0 << 2)
	#define _RGB52X_BRB_ENABLE		(1 << 2)
	#define _RGB52X_DSR_DISABLE		(0 << 1)
	#define _RGB52X_DSR_ENABLE		(1 << 1)
	#define _RGB52X_DPE_DISABLE		(0 << 0)
	#define _RGB52X_DPE_ENABLE		(1 << 0)

/*
**	Palette Control
*/
#define DAC_PAL			0x07
	#define _RGB52X_6BIT_LINEAR_ENABLE	(0 << 7)
	#define _RGB52X_6BIT_LINEAR_DISABLE	(1 << 7)
	#define _RGB52X_PALETTE_PARTITION(n)	((n) & 0xF)

#define DAC_SYSCLK		0x08	/* not RGB525 */

/*
**	Pixel Representation
*/
#define DAC_PIX_FMT		0x0a
	#define _RGB52X_PIXEL_FORMAT_4BPP	(2 << 0)
	#define _RGB52X_PIXEL_FORMAT_8BPP	(3 << 0)
	#define _RGB52X_PIXEL_FORMAT_16BPP	(4 << 0)
	#define _RGB52X_PIXEL_FORMAT_24BPP	(5 << 0)
	#define _RGB52X_PIXEL_FORMAT_31BPP	(6 << 0)

/*
**	8 Bit Pixel Control
*/
#define DAC_8BPP_CTRL		0x0b
	#define _RGB52X_B8_DCOL_INDIRECT	(0 << 0)
	#define _RGB52X_B8_DCOL_DIRECT		(1 << 0)

/*
**	16 Bit Pixel Control
*/
#define DAC_16BPP_CTRL		0x0c
	#define _RGB52X_B16_DCOL_INDIRECT	(0 << 6)
	#define _RGB52X_B16_DCOL_DYNAMIC	(1 << 6)
	#define _RGB52X_B16_DCOL_DIRECT		(3 << 6)
	#define _RGB52X_B16_POL_FORCES_BYPASS	(0 << 5)
	#define _RGB52X_B16_POL_FORCES_LOOKUP	(1 << 5)
	#define _RGB52X_B16_ZIB			(0 << 2)
	#define _RGB52X_B16_LINEAR		(1 << 2)
	#define _RGB52X_B16_555			(0 << 1)
	#define _RGB52X_B16_565			(1 << 1)
	#define _RGB52X_B16_SPARSE		(0 << 0)
	#define _RGB52X_B16_CONTIGUOUS		(1 << 0)

/*
**	24 Bit Pixel Contol
*/
#define DAC_24BPP_CTRL		0x0d
	#define _RGB52X_B24_DCOL_INDIRECT	(0 << 0)
	#define _RGB52X_B24_DCOL_DIRECT		(1 << 0)

/*
**	32 Bit Pixel Control
*/
#define DAC_32BPP_CTRL		0x0e
	#define _RGB52X_B32_POL_FORCES_BYPASS	(0 << 2)
	#define _RGB52X_B32_POL_FORCES_LOOKUP	(1 << 2)
	#define _RGB52X_B32_DCOL_INDIRECT	(0 << 0)
	#define _RGB52X_B32_DCOL_DYNAMIC	(1 << 0)
	#define _RGB52X_B32_DCOL_DIRECT		(3 << 0)

/*
**	Frequency control
*/

/*
**	PLL Control One
*/
#define DAC_PLL_CTRL1		0x10
	#define _RGB52X_REF_SRC_REFCLK		(0 << 4)
	#define _RGB52X_REF_SRC_EXTCLOCK	(1 << 4)
	#define _RGB52X_PLL_EXT_FS_DIRECT	(0 << 0)
	#define _RGB52X_PLL_EXT_FS_M_N		(1 << 0)
	#define _RGB52X_PLL_INT_FS_DIRECT	(2 << 0)
	#define _RGB52X_PLL_INT_FS_M_N		(3 << 0)

/*
**	PLL Control Two
*/
#define DAC_PLL_CTRL2		0x11
	#define _RGB52X_PLL_INT_FS(n)		((n) & 0xF)


/*
**	PLL Ref Div Count
*/
#define DAC_PLL_REF_DIV_FIX	0x14
	#define _RGB52X_REF_DIV_COUNT		((n) & 0xF)

	#define _RGB52X_PLL_REFCLK_4_MHz       	(0x02)
	#define _RGB52X_PLL_REFCLK_6_MHz       	(0x03)
	#define _RGB52X_PLL_REFCLK_8_MHz       	(0x04)
	#define _RGB52X_PLL_REFCLK_10_MHz      	(0x05)
	#define _RGB52X_PLL_REFCLK_12_MHz      	(0x06)
	#define _RGB52X_PLL_REFCLK_14_MHz      	(0x07)
	#define _RGB52X_PLL_REFCLK_16_MHz      	(0x08)
	#define _RGB52X_PLL_REFCLK_18_MHz     	(0x09)
	#define _RGB52X_PLL_REFCLK_20_MHz      	(0x0A)
	#define _RGB52X_PLL_REFCLK_22_MHz      	(0x0B)
	#define _RGB52X_PLL_REFCLK_24_MHz      	(0x0C)
	#define _RGB52X_PLL_REFCLK_26_MHz      	(0x0D)
	#define _RGB52X_PLL_REFCLK_28_MHz      	(0x0E)
	#define _RGB52X_PLL_REFCLK_30_MHz     	(0x0F)
	#define _RGB52X_PLL_REFCLK_32_MHz      	(0x10)
	#define _RGB52X_PLL_REFCLK_34_MHz      	(0x11)
	#define _RGB52X_PLL_REFCLK_36_MHz      	(0x12)
	#define _RGB52X_PLL_REFCLK_38_MHz      	(0x13)
	#define _RGB52X_PLL_REFCLK_40_MHz      	(0x14)
	#define _RGB52X_PLL_REFCLK_42_MHz      	(0x15)
	#define _RGB52X_PLL_REFCLK_44_MHz      	(0x16)
	#define _RGB52X_PLL_REFCLK_46_MHz      	(0x17)
	#define _RGB52X_PLL_REFCLK_48_MHz      	(0x18)
	#define _RGB52X_PLL_REFCLK_50_MHz      	(0x19)
	#define _RGB52X_PLL_REFCLK_52_MHz      	(0x1A)
	#define _RGB52X_PLL_REFCLK_54_MHz      	(0x1B)
	#define _RGB52X_PLL_REFCLK_56_MHz      	(0x1C)
	#define _RGB52X_PLL_REFCLK_58_MHz      	(0x1D)
	#define _RGB52X_PLL_REFCLK_60_MHz      	(0x1E)
	#define _RGB52X_PLL_REFCLK_62_MHz      	(0x1F)

#define DAC_SYSCLK_REF_DIV	0x15	/* not RGB525 */
#define DAC_SYSCLK_VCO_DIV	0x16	/* not RGB525 */

/*
**	F0-F15[7:0]
*/
#define DAC_F0			0x20
#define DAC_F1			0x21
#define DAC_F2			0x22
#define DAC_F3			0x23
#define DAC_F4			0x24
#define DAC_F5			0x25
#define DAC_F6			0x26
#define DAC_F7			0x27
#define DAC_F8			0x28
#define DAC_F9			0x29
#define DAC_F10			0x2A
#define DAC_F11			0x2B
#define DAC_F12			0x2C
#define DAC_F13			0x2D
#define DAC_F14			0x2E
#define DAC_F15			0x2F

	#define _RGB52X_DF(n)			(((n) & 0x3) << 6)
	#define _RGB52X_VCO_DIV_COUNT(n)	((n) & 0x3F)

#define DAC_M0			0x20
#define DAC_M1			0x22
#define DAC_M2			0x24
#define DAC_M3			0x26
#define DAC_M4			0x28
#define DAC_M5			0x2A
#define DAC_M6			0x2C
#define DAC_M7			0x2E

#define DAC_N0			0x21
#define DAC_N1			0x23
#define DAC_N2			0x25
#define DAC_N3			0x27
#define DAC_N4 			0x29
#define DAC_N5  		0x2B
#define DAC_N6			0x2D
#define DAC_N7			0x2F


#define DAC_CURS_CTRL		0x30
	#define _RGB52X_SMLC_PART0			(0 << 6)
	#define _RGB52X_SMLC_PART1			(1 << 6)
	#define _RGB52X_SMLC_PART2			(2 << 6)
	#define _RGB52X_SMLC_PART3			(3 << 6)
	#define _RGB52X_PIX_ORDER_R_TO_L		(0 << 5)
	#define _RGB52X_PIX_ORDER_L_TO_R		(1 << 5)
	#define _RGB52X_LOC_READ_LAST_WRITTEN		(0 << 4)
	#define _RGB52X_LOC_READ_ACTUAL_LOCATION	(1 << 4)
	#define _RGB52X_UPDT_CTRL_DELAYED		(0 << 3)
	#define _RGB52X_UPDT_CTRL_IMMEDIATE		(1 << 3)
	#define _RGB52X_CURSOR_SZE_32			(0 << 2)
	#define _RGB52X_CURSOR_SIZE_64			(1 << 2)
	#define _RGB52X_CURSOR_MODE_OFF			(0 << 0)
	#define _RGB52X_CURSOR_MODE_3_COLOR		(1 << 0)
	#define _RGB52X_CURSOR_MODE_2_COLOR_HL		(2 << 0)
	#define _RGB52X_CURSOR_MODE_2_COLOR		(3 << 0)



#define DAC_CURS_XLOW		0x31
#define DAC_CURS_XHIGH		0x32
#define DAC_CURS_YLOW		0x33
#define DAC_CURS_YHIGH		0x34
#define DAC_CURS_HOTSPOT_X     	0x35
#define DAC_CURS_HOTSPOT_Y     	0x36

#define DAC_CURS_COL1_R		0x40
#define DAC_CURS_COL1_G		0x41
#define DAC_CURS_COL1_B		0x42
#define DAC_CURS_COL2_R		0x43
#define DAC_CURS_COL2_G		0x44
#define DAC_CURS_COL2_B		0x45
#define DAC_CURS_COL3_R		0x46
#define DAC_CURS_COL3_G		0x47
#define DAC_CURS_COL3_B		0x48

#define DAC_BORDER_COL_R	0x60
#define DAC_BORDER_COL_G	0x61
#define DAC_BORDER_COL_B	0x62

/*
**	Misc Control One
*/
#define DAC_MISC_CTRL_1		0x70
	#define _RGB52X_MISR_CTRL_OFF		(0 << 7)
	#define _RGB52X_MISR_CTRL_ON		(1 << 7)
	#define _RGB52X_VMSK_CTRL_OFF		(0 << 6)
	#define _RGB52X_VMSK_CTRL_ON		(1 << 6)
	#define _RGB52X_PADR_RFMT_READ_ADDR    	(0 << 5)
	#define _RGB52X_PADR_RFMT_PAL_STATE    	(1 << 5)
	#define _RGB52X_SENS_DSAB_ENADLE       	(0 << 4)
	#define _RGB52X_SENS_DSAB_DISABLE      	(1 << 4)
	#define _RGB52X_SENS_SEL_BIT3		(0 << 3)
	#define _RGB52X_SENS_SEL_BIT7		(1 << 3)
	#define _RGB52X_VRAM_SIZE_32		(0 << 0)
	#define _RGB52X_VRAM_SIZE_64		(1 << 0)

/*
**	Misc Control Two
*/
#define DAC_MISC_CTRL_2		0x71
	#define _RGB52X_PCLK_SEL_LCLK		(0 << 6)
	#define _RGB52X_PCLK_SEL_PLL		(1 << 6)
	#define _RGB52X_PCLK_SEL_EXT		(2 << 6)
	#define _RGB52X_INTL_MODE_DISABLE	(0 << 5)
	#define _RGB52X_INTL_MODE_ENABLE	(1 << 5)
	#define _RGB52X_BLANK_CTRL_NORMAL	(0 << 4)
	#define _RGB52X_BLANK_CTRL_BLANKED	(1 << 4)
	#define _RGB52X_COL_RES_6BIT		(0 << 2)
	#define _RGB52X_COL_RES_8BIT		(1 << 2)
	#define _RGB52X_PORT_SEL_VGA		(0 << 0)
	#define _RGB52X_PORT_SEL_VRAM		(1 << 0)

/*
**	Misc Control Three
*/
#define DAC_MISC_CTRL_3		0x72
	#define _RGB52X_SWAP_RB_DISABLE		(0 << 7)
	#define _RGB52X_SWAP_RB_ENABLE		(1 << 7)
	#define _RGB52X_SWAP_WORD_31_00_FIRST	(0 << 4)
	#define _RGB52X_SWAP_WORD_63_32_FIRST	(1 << 4)
	#define _RGB52X_SWAP_NIB_07_04_FIRST	(0 << 2)
	#define _RGB52X_SWAP_NIB_03_00_FIRST	(1 << 2)

/*
**	Misc Diagnostic Support
*/
#define DAC_DAC_SENSE		0x82
#define DAC_MISR_R		0x84
#define DAC_MISR_G		0x86
#define DAC_MISR_B		0x88
#define DAC_PLL_VCO_DIV_IN	0x8e
#define DAC_PLL_REF_DIV_IN	0x8f
#define DAC_VRAM_MASK_0		0x90
#define DAC_VRAM_MASK_1		0x91
#define DAC_VRAM_MASK_2		0x92
#define DAC_VRAM_MASK_3		0x93

#define DAC_CURS_ARRAY		0x100


#endif /* #ifdef _IBM_RGB526_H */
