/* ---------------------------------------------------------------------------
**	PS/2 auxiliary keyboard device driver
** ---------------------------------------------------------------------------
**
**	Copyright (C) 1997	Steffen Seeger
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** -------------------------------------------------------------------------
** 
**	$Log: ps2kbd.c,v $
**	Revision 1.1  1997/12/31 16:20:24  seeger
**	Merged keyboard.c and ps2aux.c. However, this is still alpha.
**	
*/

#include <ggi/maintainers.h>
#define	MAINTAINER	Steffen_Seeger

#define DRIVER_NAME	"PS/2 kbd/aux driver"
#define DRIVER_REV	"$Revision: 1.1 $"
#define	DEBUG_LEVEL	1

#include <linux/config.h>
#include <linux/module.h>

#include <ggi/kgi.h>

#include <linux/random.h>

#include "ps2kbd.h"

#if defined(__alpha__) && !defined(CONFIG_PCI)
#	define	AUX_PTR_IRQ	9	/* Jensen is odd indeed */
#	define	AUX_KBD_IRQ	1
#else
#	define	AUX_PTR_IRQ	12
#	define	AUX_KBD_IRQ	1
#endif

#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,37)
#	define	test_and_clear_bit(x,y)	clear_bit(x,y)
#	define	test_and_set_bit(x,y)	set_bit(x,y)
#endif

/*
** !!!	This is a quick backward compatibility hack!
*/
int psaux_read_mask = 0;


unsigned int aux_ptr_irq = AUX_PTR_IRQ;
unsigned int aux_kbd_irq = AUX_KBD_IRQ;
int aux_ptr_focus = -1;
int aux_kbd_focus = -1;

#define	AUX_TIMEOUT_CONST	0x20000

/*
**	misc I/O and helper functions
*/
static void aux_out(uint8 val)
{
	int i = AUX_TIMEOUT_CONST;

	while ((io_in8(AUX_STATUS) & AUX_STATUS_OBF) && --i) {
	}

	if (i) {

		DEBUG2("aux_out <- %.2x", val);
		io_out8(val, AUX_OUT);

	} else {

		DEBUG2("aux_out timed out");
	}
}

static uint8 aux_in(void)
{
	int i = AUX_TIMEOUT_CONST;

	while ((~io_in8(AUX_STATUS) & AUX_STATUS_IBF) && --i) {
	}

	if (i) {

		uint8 foo = io_in8(AUX_IN);
		DEBUG2("aux_in -> %.2x, i == %i", foo, i);
		return foo;

	} else {

		DEBUG2("aux_in timed out");
		return 0xFF;
	}
}

static inline void aux_cmd(uint8 cmd)
{
	int i = AUX_TIMEOUT_CONST;

	while ((io_in8(AUX_STATUS) & AUX_STATUS_OBF) && --i) {
	}

	if (i) {

		DEBUG2("aux_cmd <- %.2x", cmd);
		io_out8(cmd, AUX_CMD);

	} else {

		DEBUG2("aux_cmd timed out");
	}
}

static inline void aux_flush(void)
{
	int i;

	i = AUX_TIMEOUT_CONST;
	do {

		while ((~io_in8(AUX_STATUS) & AUX_STATUS_IBF) && --i) {
		}

		if (i) {

			i = AUX_TIMEOUT_CONST;
			io_in8(AUX_IN);
		} 

	} while (i);
}

static inline uint8 aux_kbd_send(uint8 dat)
/*	Send a datum to the device connected to the kbd port.
*/
{
	int i = 3;
	uint8 reply;

	do {
		aux_out(dat);
		reply = aux_in();

	} while ((reply == DEV_RESEND) && i--);

	return reply;
}

static inline uint8 aux_ptr_send(uint8 dat)
/*	Send a datum to the device connected to the ptr port.
*/
{
	uint8 reply;
	int i = 3;

	do {
		aux_cmd(AUX_CMD_PTR_WRITE);
		aux_out(dat);
		reply = aux_in();

	} while ((reply == DEV_RESEND) && i--);

	return reply;
}

/*
**	the interrupt handlers
*/

static void *aux_kbd_parser_priv = NULL;
static void *aux_ptr_parser_priv = NULL;
static void (*aux_kbd_parser)(void *priv, uint8 data) = NULL;
static void (*aux_ptr_parser)(void *priv, uint8 data) = NULL;

static void aux_kbd_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	uint8 status, data;
	aux_cmd(AUX_CMD_KBD_DISABLE);

	status = io_in8(AUX_STATUS);
	if ((status & AUX_STATUS_OBT) == 0) {

		data = io_in8(AUX_IN);	

		if (aux_kbd_parser) {

			aux_kbd_parser(aux_kbd_parser_priv, data);
		}
	}

	aux_cmd(AUX_CMD_KBD_ENABLE);
}

static void aux_ptr_interrupt(int irq, void *dev_id, struct pt_regs *pt_regs)
{
	uint8 data;

	aux_cmd(AUX_CMD_PTR_DISABLE);

	if (io_in8(AUX_STATUS) & AUX_STATUS_OBT) {

		data = io_in8(AUX_IN);

		if (aux_ptr_parser) {
	
			aux_ptr_parser(aux_ptr_parser_priv, data);
		}
	}

	aux_cmd(AUX_CMD_PTR_ENABLE);
}


static struct io_region aux_io_region =
{
	IO_NULL,	0x60,	16,	0xFFFF,	"kbd/aux interface"
};

#define	irq_check	check_irq
#define	irq_claim	request_irq

static int aux_ptr = 0, aux_kbd = 0;

/*
**	keyboard  parser
*/

#define	BYTES(x)	(((x) + 7) & ~7)
#define	MAX_NR_KEYS	128

struct aux_parser_kbd_data
{
	struct	kgi_input	input;

	uint	prev_scancode;
	uint8	keys[BYTES(MAX_NR_KEYS)];	
};

/*	scancode to keycode translation
**
**	This is now user-settable. The keycodes 1-88,96-111,119 are fairly
**	standard, and should probably not be changed - changing might confuse X.
**	X also interprets scancode 0x5d (KEY_Begin). For 1-88 keycode equals
**	scancode.
*/

#define E0_KPENTER 96
#define E0_RCTRL   97
#define E0_KPSLASH 98
#define E0_PRSCR   99
#define E0_RALT    100
#define E0_BREAK   101  /* (control-pause) */
#define E0_HOME    102
#define E0_UP      103
#define E0_PGUP    104
#define E0_LEFT    105
#define E0_RIGHT   106
#define E0_END     107
#define E0_DOWN    108
#define E0_PGDN    109
#define E0_INS     110
#define E0_DEL     111

#define E1_PAUSE   119

/*	The keycodes below are randomly located in 89-95,112-118,120-127.
**	They could be thrown away (and all occurrences below replaced by 0),
**	but that would force many users to use the `setkeycodes' utility, where
**	they needed not before. It does not matter that there are duplicates, as
**	long as no duplication occurs for any single keyboard.
*/

#define SC_LIM 89

#define FOCUS_PF1 85           /* actual code! */
#define FOCUS_PF2 89
#define FOCUS_PF3 90
#define FOCUS_PF4 91
#define FOCUS_PF5 92
#define FOCUS_PF6 93
#define FOCUS_PF7 94
#define FOCUS_PF8 95
#define FOCUS_PF9 120
#define FOCUS_PF10 121
#define FOCUS_PF11 122
#define FOCUS_PF12 123

#define JAP_86     124

/*	tfj@olivia.ping.dk: The four keys are located over the numeric 
**	keypad, and are labelled A1-A4. It's an rc930 keyboard, from
**	Regnecentralen/RC International, Now ICL. Scancodes: 59, 5a, 5b, 5c.
*/

#define RGN1 124
#define RGN2 125
#define RGN3 126
#define RGN4 127


/* BTC */
#define E0_MACRO   112
/* LK450 */
#define E0_F13     113
#define E0_F14     114
#define E0_HELP    115
#define E0_DO      116
#define E0_F17     117
#define E0_KPMINPLUS 118

/*	kkoller@nyx10.cs.du.edu: My OmniKey generates e0 4c for the "OMNI" 
**	key and the right alt key does nada.
*/

#define E0_OK	124

/*	New microsoft keyboard is rumoured to have
**	e0 5b (left window button),	LBANNER or Windows_L
**	e0 5c (right window button),	RBANNER or Windows_R
**	e0 5d (menu button).		RMENU   or TaskMan
*/

#define E0_MSLW	125
#define E0_MSRW	126
#define E0_MSTM	127

static uint8 high_keys[128 - SC_LIM] = {

	RGN1, RGN2, RGN3, RGN4, 0, 0, 0,                   /* 0x59-0x5f */
	0, 0, 0, 0, 0, 0, 0, 0,                            /* 0x60-0x67 */
	0, 0, 0, 0, 0, FOCUS_PF11, 0, FOCUS_PF12,          /* 0x68-0x6f */
	0, 0, 0, FOCUS_PF2, FOCUS_PF9, 0, 0, FOCUS_PF3,    /* 0x70-0x77 */
	FOCUS_PF4, FOCUS_PF5, FOCUS_PF6, FOCUS_PF7,        /* 0x78-0x7b */
	FOCUS_PF8, JAP_86, FOCUS_PF10, 0                   /* 0x7c-0x7f */
};

static uint8 e0_keys[128] = {

	0, 0, 0, 0, 0, 0, 0, 0,				/* 0x00-0x07 */
	0, 0, 0, 0, 0, 0, 0, 0,				/* 0x08-0x0f */
	0, 0, 0, 0, 0, 0, 0, 0,				/* 0x10-0x17 */
	0, 0, 0, 0, E0_KPENTER, E0_RCTRL, 0, 0,		/* 0x18-0x1f */
	0, 0, 0, 0, 0, 0, 0, 0,				/* 0x20-0x27 */
	0, 0, 0, 0, 0, 0, 0, 0,				/* 0x28-0x2f */
	0, 0, 0, 0, 0, E0_KPSLASH, 0, E0_PRSCR,		/* 0x30-0x37 */
	E0_RALT, 0, 0, 0, 0, E0_F13, E0_F14, E0_HELP,	/* 0x38-0x3f */
	E0_DO, E0_F17, 0, 0, 0, 0, E0_BREAK, E0_HOME,	/* 0x40-0x47 */
	E0_UP, E0_PGUP, 0, E0_LEFT, E0_OK, E0_RIGHT, E0_KPMINPLUS, E0_END,/* 0x48-0x4f */
	E0_DOWN, E0_PGDN, E0_INS, E0_DEL, 0, 0, 0, 0,	/* 0x50-0x57 */
	0, 0, 0, E0_MSLW, E0_MSRW, E0_MSTM, 0, 0,	/* 0x58-0x5f */
	0, 0, 0, 0, 0, 0, 0, 0,				/* 0x60-0x67 */
	0, 0, 0, 0, 0, 0, 0, E0_MACRO,			/* 0x68-0x6f */
	0, 0, 0, 0, 0, 0, 0, 0,				/* 0x70-0x77 */
	0, 0, 0, 0, 0, 0, 0, 0				/* 0x78-0x7f */
};

#if 0
static int kbd_set_keycode(ggi_uint scancode, ggi_uint keycode)
{
	if ((scancode < SC_LIM) || (scancode > 255) || (keycode > 127)) {

		return -EINVAL;
	}

	if (scancode < 128) {

		high_keys[scancode - SC_LIM] = keycode;

	} else {

		e0_keys[scancode - 128] = keycode;
	}

	return EOK;
}

static int kbd_get_keycode(ggi_uint scancode)
{
	return ((scancode < SC_LIM) || (scancode > 255))
			? -EINVAL
			: (scancode < 128)
				? high_keys[scancode - SC_LIM]
				: e0_keys[scancode - 128];
}
#endif

static void aux_parser_keyboard(void *priv, uint8 scancode)
{
	struct aux_parser_kbd_data *kbd = (struct aux_parser_kbd_data *) priv;
	ggi_event event;
	uint8 release;
	uint keycode;

	add_keyboard_randomness(scancode);

	switch (scancode) {

	case 0:
		ERROR("keyboard buffer overflow");
		return;

	case DEV_ACKNOWLEDGE:
	case DEV_RESEND:
		return;

	case 0xFF:
		ERROR("keyboard error");
		return;
	}

	event.any.focus  = kbd->input.focus;
	event.any.device = kbd->input.id;
	event.any.time   = jiffies;

	if (kbd->input.report & emRawData) {

		event.raw.type	  = evRawData;
		event.raw.size	  = sizeof(ggi_any_event) + 1;
		event.raw.data[0] = scancode;

		kgi_handle_input(&event);
	}

	if (scancode == 0xE0 || scancode == 0xE1) {

		kbd->prev_scancode = scancode;
		return;
	}

	release = scancode & 0x80;
	scancode &= 0x7F;

	if (kbd->prev_scancode) {

		/*	usually it will be 0xe0, but a Pause key generates
		**	e1 1d 45 e1 9d c5 when pressed, and nothing when
		**	released.
		*/
		if (kbd->prev_scancode != 0xe0) {

			if (kbd->prev_scancode == 0xe1 && scancode == 0x1d) {

				kbd->prev_scancode = 0x100;
				return;
	
			} else {

				if (kbd->prev_scancode == 0x100 &&
					scancode == 0x45) {
	
					keycode = E1_PAUSE;
					kbd->prev_scancode = 0;

				} else {


					ERROR("unknown E1 sequence");
					kbd->prev_scancode = 0;
					return;
				}
			}

		} else {	/* if (kbd->prev_scancode != 0xe0) { ... */

			kbd->prev_scancode = 0;

			/*	The keyboard maintains its own internal capsi
			**	lock and num lock statuses. In caps lock mode
			**	E0 AA precedes make code and E0 2A follows
			**	break code. In num lock mode, E0 2A precedes
			**	make code and E0 AA follows break code. We do
			**	our own book-keeping, so we just ignore these.
			**
			**	For my keyboard there is no caps lock mode,
			**	but there are both Shift-L and Shift-R modes.
			**	The former mode generates E0 2A / E0 AA pairs,
			**	the latter E0 B6 / E0 36 pairs. So, we should
			**	also ignore the latter. -- aeb@cwi.nl
			*/
			if (scancode == 0x2a || scancode == 0x36) {

				return;
			}

			if (e0_keys[scancode]) {

				keycode = e0_keys[scancode];

			} else {

				ERROR("unknown scancode e0 %02x", scancode);
				return;
			}
		}

	} else {  /* if (kbd->prev_scancode) ... */

		if (scancode >= SC_LIM) {

		        /*	This happens with the FOCUS 9000 keyboard.
			**	Its keys PF1..PF12 are reported to generate
			**	55 73 77 78 79 7a 7b 7c 74 7e 6d 6f. Moreover,
			**	unless repeated, they do not generate key-down
			**	events, so we have to zero <release> below.
			**
			**	Also, Japanese 86/106 keyboards are reported
			**	to generate 0x73 and 0x7d for \ - and \ |
			**	respectively.  
			**
			**	Also, some Brazilian keyboard is reported to
			**	produce 0x73 and 0x7e for \ ? and KP-dot,
			**	respectively.
			*/

			keycode = high_keys[scancode - SC_LIM];

			if (! keycode) {

				ERROR("unknown scancode %2x", scancode);
				return;
			}

		} else {

			 keycode = scancode;
		}
	}

	event.key.size = sizeof(ggi_key_event);
	event.key.code = keycode;

	if (release) {

		event.key.type = evKeyRelease;
		if (! test_and_clear_bit(keycode, kbd->keys)) {

			/*	unexpected, but this can happen: maybe this
			**	was a key release for a FOCUS 9000 PF key;
			*/
			if (keycode >= SC_LIM || keycode == 85) {

				event.key.type = evKeyPress;
				if (kbd->input.report & emKeyPress) {

					kgi_handle_input(&event);
				}
				event.key.type = evKeyRelease;
			}
		}

	} else {

		event.key.type = test_and_set_bit(keycode, kbd->keys)
			? evKeyRepeat : evKeyPress;
	}

	if (kbd->input.report & (1 << event.key.type)) {

		kgi_handle_input(&event);
	}
}

/*
**	pointer parser
*/

struct aux_parser_ptr_data
{
	struct kgi_input	input;

	ggi_uint		state;
	ggi_uint		b, last_b;
	ggi_sint		dx, dy;
};

static void aux_parser_pointer(void *priv, uint8 data)
{
	struct aux_parser_ptr_data *ptr = (struct aux_parser_ptr_data *) priv;
	ggi_event event;

	add_mouse_randomness(data);

	switch (ptr->state) {

	case 0:
		ptr->b = data;
		ptr->state++;
		return;

	case 1:
		ptr->dx = data;
		if (ptr->b & 0x10) {

			ptr->dx |= (-1 & ~0xFF);
		}
		ptr->state++;
		return;

	case 2:
		ptr->dy = data;
		if (ptr->b & 0x20) {

			ptr->dy |= (-1 & ~0xFF);
		}
		ptr->state = 0;
		ptr->b &= 7;
	}


	DEBUG2("packet %.2x %i, %i", ptr->b, ptr->dx, ptr->dy);
	/* !!!	do mouse acceleration here !!! */

	event.any.focus  = ptr->input.focus;
	event.any.device = ptr->input.id;
	event.any.time   = jiffies;

	if (ptr->dx || ptr->dy) {

		event.pmove.size = sizeof(ggi_pmove_event);

		event.pmove.wheel = 0;
		event.pmove.dummy[0] = event.pmove.dummy[1] = 0;

		event.pmove.x =  ptr->dx;
		event.pmove.y = -ptr->dy;

		if (ptr->input.report & emPtrRelative) {

			DEBUG2("evPtrRelative");
			event.pmove.type = evPtrRelative;
			kgi_handle_input(&event);
		}
	}

	/*	Change in button state?
	*/
	if ((event.pbutton.button = ptr->last_b ^ ptr->b)) {

		event.pbutton.size = sizeof(ggi_pbutton_event);
		ptr->last_b = event.pbutton.state = ptr->b;
		event.pbutton.type =
			(event.pbutton.button & event.pbutton.state)
			? evPtrButtonPress : evPtrButtonRelease;

		if (ptr->input.report & (1 << event.pbutton.type)) {

			DEBUG2("%s", 
			       (event.pbutton.type == evPtrButtonPress)
			       ? "evPtrButtonPress" 
			       : "evPtrButtonRelease");
			kgi_handle_input(&event);
		}
	}	
}

typedef union {

	struct kgi_input	input;
	struct aux_parser_kbd_data	kbd;
	struct aux_parser_ptr_data	ptr;

} aux_private;

aux_private aux_ptr_priv, aux_kbd_priv;

#define	SIG_KEYBOARD	0x83ABFA
#define	SIG_MOUSE	0xFFFF00

static int aux_init_input(aux_private *priv, ggi_uint sig, char *port,
	ggi_sint focus)
{
	memset(priv, 0, sizeof(*priv));
	priv->input.focus	= -1;
	priv->input.id	= 1;
	strcpy (priv->input.manufact, "IBM");

	switch (sig) {

	case SIG_KEYBOARD:
		priv->input.events = emKey | emRawData;
		priv->input.report = emKey;
		sprintf(priv->input.model, "PS/2 keyboard (%s)", port);
		break;

	case SIG_MOUSE:
		priv->input.events = emPointer & ~emPtrAbsolute;
		priv->input.report = emPtrRelative | emPtrButton;
		sprintf(priv->input.model, "PS/2 mouse (%s)", port);
		break;
	default:
		return -EINVAL;
	}

	return kgi_register_input(focus, &(priv->input));
}


#ifndef	MODULE
static int aux_init_devices(void)
#else
int init_module(void)
#endif
{
	int	signature;

	/*	claim region, disable interrupts and flush any pending input.
	*/
	if (io_check_region(&aux_io_region)) {

		DEBUG0("aux_io_region busy");
		return -EBUSY;
	}
	io_claim_region(&aux_io_region);

	aux_cmd(AUX_CMD_WRITE_MODE);
	aux_out(0);
	aux_flush();

	/*	do controller self test.
	*/
	aux_cmd(AUX_CMD_SELF_TEST);
	if (aux_in() != 0x55) {

		ERROR("controller self test failed");
		io_free_region(&aux_io_region);
		return -ENODEV;
	}

	/*	test kbd and ptr interface and claim interrupts
	*/
	aux_cmd(AUX_CMD_KBD_TEST);
	aux_kbd = (aux_in() == 0) &&
		(request_irq(aux_kbd_irq, aux_kbd_interrupt, 0, "aux kbd",
		NULL) == 0);

	aux_cmd(AUX_CMD_PTR_TEST);
	aux_ptr = (aux_in() == 0) &&
		(request_irq(aux_ptr_irq, aux_ptr_interrupt, 0, "aux ptr",
		NULL) == 0);

	aux_cmd(AUX_CMD_PTR_DISABLE);
	aux_cmd(AUX_CMD_KBD_DISABLE);

	if (! (aux_ptr || aux_kbd)) {

		ERROR("no devices detected");
		io_free_region(&aux_io_region);
		return -ENODEV;
	}

	if (aux_ptr) {

		aux_cmd(AUX_CMD_PTR_ENABLE);

		aux_ptr_send(DEV_CMD_RESET);
		signature =  aux_ptr_send(KBD_CMD_ID);
		signature |= aux_in() << 8;
		signature |= aux_in() << 16;

		switch (signature) {

		case SIG_MOUSE:
			DEBUG0("mouse detected on aux ptr port");
			aux_ptr_send(DEV_CMD_ENABLE);
			aux_ptr_parser = aux_parser_pointer;
			if (aux_ptr_focus < 0) {

				aux_ptr_focus = 0;
			}
			break;

		case SIG_KEYBOARD:
			DEBUG0("keyboard detected on aux ptr port");
			aux_ptr_send(KBD_CMD_SCANCODE);
			aux_ptr_send(KBD_SC_1);
			aux_ptr_send(DEV_CMD_ENABLE);
			if (aux_ptr_focus < 0) {

				aux_ptr_focus = 1;
			}
			aux_ptr_parser = aux_parser_keyboard;
			break;

		default:
			DEBUG0("unknown signature %.6x from ptr", signature);
			free_irq(aux_ptr_irq, NULL);
			aux_ptr = 0;
		}

		if (aux_ptr && (aux_init_input(&aux_ptr_priv, signature,
			"ptr", aux_ptr_focus) != EOK)) {

			DEBUG0("%s %s registration failed",
				aux_ptr_priv.input.manufact,
				aux_ptr_priv.input.model);
			free_irq(aux_ptr_irq, NULL);
			aux_ptr = 0;
		}

		aux_cmd(AUX_CMD_PTR_DISABLE);
	}

	if (aux_kbd) {

		aux_cmd(AUX_CMD_KBD_ENABLE);

		aux_kbd_send(DEV_CMD_RESET);
		signature =  aux_kbd_send(KBD_CMD_ID);
		signature |= aux_in() << 8;
		signature |= aux_in() << 16;

		switch (signature) {

		case SIG_MOUSE:
			DEBUG0("mouse detected on aux_kbd port");
			aux_kbd_send(DEV_CMD_ENABLE);
			aux_kbd_parser = aux_parser_pointer;
			if (aux_kbd_focus < 0) {

				aux_kbd_focus = 0;
			}
			break;

		case SIG_KEYBOARD:
			DEBUG0("keyboard detected on aux kbd port");
			aux_kbd_send(KBD_CMD_SCANCODE);
			aux_kbd_send(KBD_SC_1);
			aux_kbd_send(DEV_CMD_ENABLE);
			aux_kbd_parser = aux_parser_keyboard;
			if (aux_kbd_focus < 0) {

				aux_kbd_focus = 0;
			}
			break;

		default:
			DEBUG0("unknown signature %.6x from kbd", signature);
			aux_cmd(AUX_CMD_KBD_DISABLE);
			aux_cmd(AUX_CMD_WRITE_MODE);
			free_irq(aux_kbd_irq, NULL);
			aux_kbd = 0;
		}

		if (aux_kbd && (aux_init_input(&aux_kbd_priv, signature,
			"kbd", aux_kbd_focus) != EOK)) {

			DEBUG0("%s %s registration failed",
				aux_kbd_priv.input.manufact,
				aux_kbd_priv.input.model);
			free_irq(aux_ptr_irq, NULL);
			aux_ptr = 0;
			aux_kbd_parser = NULL;
		}
	}

	if (! (aux_ptr || aux_kbd)) {

		ERROR("not loaded: only unknown devices");
		io_free_region(&aux_io_region);
		return -ENODEV;
	}

	if (aux_kbd) {

		aux_cmd(AUX_CMD_KBD_ENABLE);
		aux_kbd_parser_priv = &aux_kbd_priv;
	}
	if (aux_ptr) {

		aux_cmd(AUX_CMD_PTR_ENABLE);
		aux_ptr_parser_priv = &aux_ptr_priv;
	}
	aux_cmd(AUX_CMD_WRITE_MODE);
	aux_out((aux_kbd ? AUX_MODE_KBD_INTS_ON : 0) |
		(aux_ptr ? AUX_MODE_PTR_INTS_ON : 0));

	return EOK;
}

#ifdef	MODULE
void cleanup_module(void)
{
	aux_cmd(AUX_CMD_WRITE_MODE);
	aux_out(0);

	if (aux_kbd) {

		free_irq(aux_kbd_irq, NULL);
	}
	if (aux_ptr) {

		free_irq(aux_ptr_irq, NULL);
	}
	io_free_region(&aux_io_region);
	DEBUG0("driver unloaded");
}
#else	/* #ifdef MODULE	*/

int kbd_init(void)
{
	int keyboards = 0;

	if (aux_init_devices() == EOK) {

		if (aux_kbd_priv.input.events & emKey) {

			keyboards++;
		}
		if (aux_ptr_priv.input.events & emKey) {

			keyboards++;
		}
	}
	printk("%i keyboards detected\n", keyboards);

	return keyboards;
}

#endif	/* #ifdef MODULE	*/
