/* ---------------------------------------------------------------------------
**	Cirrus Logic GD546x (Laguna) ramdac driver
** ---------------------------------------------------------------------------
**
**	Copyright (C) 1997 Rodolphe Ortalo
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** ---------------------------------------------------------------------------
**
**	$Log: 546x.c,v $
**	Revision 1.5  1997/12/18 22:38:34  becka
**	Fixed up RAMDAC drivers. Small spelling error corrected.
**	
**	Revision 1.4  1997/12/17 21:43:45  ortalo
**	Removed the palette initialization (ggi/palette/ *.inc files
**	have disappeared...).
**	
**	Revision 1.3  1997/12/14 21:56:18  becka
**	uint/sint namespace fixed ... please test ...
**	
**	Revision 1.2  1997/10/12 17:56:32  ortalo
**	Ramdac module okay for some graphic modes (and font setting). Hardware
**	cursor support added (but not tested).
**	
**	Revision 1.1  1997/09/14 19:15:03  ortalo
**	New ramdac module for the 546x chipset. Based on the 542x. get/set_clut
**	functions are provided.
**	
**
** ---------------------------------------------------------------------------
*/

#include <ggi/maintainers.h>
#define	MAINTAINER	"Rodolphe Ortalo, [ortalo@laas.fr]"
#define DRIVER_NAME	"Cirrus Logic CL-GD546x (Laguna) RamDAC driver"
#define DRIVER_REV	"$Revision: 1.5 $"

#define DEBUG 2
#ifdef DEBUG
#define DEBUG_LEVEL DEBUG
#endif

#ifdef DEBUG
#define	NEED_GRAPHTYPE_SYMS
#endif

#include <ggi/kgi.h>
#include <ggi/module.h>
#include <ggi/ggi_commands.h>

#include "Cirrus/546x.h"

/* Ugly ? -- ortalo */
extern int CLIO_in8(int);
extern void CLIO_out8(int,int);
extern int CLIO_in16(int);
extern void CLIO_out16(int,int);
extern struct mem_region linFrameBuffer;

/*
**	Ramdac description
*/

struct kgi_ramdac 
ramdac = {	"Cirrus Logic",	       		/* manufacturer		*/
	 	"CL-GD546x (Laguna)",		/* model		*/
 		GM_TEXT16 | GM_1BIT | GM_4BIT |	/* supported modes	*/
		GM_8BIT | GM_15BIT | GM_16BIT | GM_24BIT | GM_32BIT ,
	 	0,			 	/* flags		*/
	 	{ 0xFFFF, 0xFFFF },		/* maximum resolution	*/
	#warning RAMDAC max freq is different between 5462 and 5464: update !
		 DAC_8BIT_PB, 			/* device type		*/
		  {       0,  170000000 },	/* normal mode		*/
		  {       0,         0 },	/* 2xclk mode		*/

		 {{	  0,  170000000 },	/* 	GT_TEXT16	*/
		  {	  0,         0 },	/* 	GT_TEXT32 	*/
		  {       0,  170000000 },	/* 	GT_1BIT		*/
		  {	  0,  170000000 },	/* 	GT_4BIT		*/
/*		  {       0,  170000000 },*/	/* 	GT_8BIT		*/
		  {       0,  170000000 },	/* 	GT_8BIT		*/
		  {	  0,  170000000 },	/* 	GT_15BIT	*/
		  { 	  0,  80000000 },	/* 	GT_16BIT	*/
		  {       0,  170000000 },	/* 	GT_24BIT	*/
		  {	  0,  170000000 }}	/* 	GT_32BIT	*/
         };

/*
** The following parameters indicate the transformations
** we must do before setting a color in the CLUT.
*/
 #define SHIFT 8
#define MAGIC 257

/*
**	driver state info
*/

struct dac_private 
{
	ggi_uint	cur_hotx, cur_hoty;	/* cursor hot spot	*/
	ggi_uint	cur_x, cur_y;		/* current position	*/
	ggi_uint	cur_enabled;		/* enabled		*/
};

#define	CUR_HOTX	private->cur_hotx
#define	CUR_HOTY	private->cur_hoty
#define	CUR_X		private->cur_x
#define	CUR_Y		private->cur_y
#define CUR_ENABLED	private->cur_enabled

#define	GET_PRIVATE	struct dac_private *private = KGIM_PRIVATE(dpy,ramdac)

/*
**	Hardware pointer functions
*/


static void dac_set_pointer_shape(struct kgi_display *dpy, 
	struct kgi_pointer64x64 *ptr)
{
	GET_PRIVATE;
	ggi_uint i,j;
	uint8 *addr;
	ggi_uint fb_addr;

	switch (ptr->mode) {
		case pmWindows:
		case pmX11:
		case pmThreeColor:
			NOTICE2("Legal cursor mode - but same !");
			break;
		default:
			INTERNAL_ERROR;
	}

	/*	Setting hot spot
	*/
	CUR_HOTX = ptr->hotx;
	CUR_HOTY = ptr->hoty;
	CLIO_out16(CURSOR_PRESET, ((ptr->hotx << 8) & CURSOR_X_PRESET)
				| ( ptr->hoty & CURSOR_Y_PRESET) );

	/* Storing cursor data at the end of the framebuffer
	*/
	fb_addr = (dpy->card->chipset->ramsize - 1024) & ~0x3FF;
	addr = linFrameBuffer.baseptr + fb_addr;
	CLIO_out16(CURSOR_ADDRESS, (fb_addr >> 8) & CURSOR_ADDRESS_MASK);

	DEBUG2("set_pointer_shape(64x64): hot spot (%i,%i) fb_addr:%d addr:%p",
	       CUR_HOTX, CUR_HOTY, fb_addr, addr);

	/*	set pattern (interleaved 64 bits quad-words)
	*/
	for (i = 0; i < 512; i+=8) {
		for (j = 0; j < 8; j++)
			(*addr++) = ptr->xor_mask[i+j];
		for (j = 0; j < 8; j++)	
			(*addr++) = ~(ptr->and_mask[i+j]);
	}

	/*	set colors
	*/
	CLIO_out8(PALETTE_STATE,
		CLIO_in8(PALETTE_STATE) | ACCESS_CURSOR_COLORS);
	/* cursor color 0 */
	CLIO_out8(DAC_PALETTE_ADDRESS_WRITE, 0x00);
	CLIO_out8(DAC_PIXEL_DATA, ptr->col[0].r >> SHIFT); /* red */
	CLIO_out8(DAC_PIXEL_DATA, ptr->col[0].g >> SHIFT); /* green */
	CLIO_out8(DAC_PIXEL_DATA, ptr->col[0].b >> SHIFT); /* blue */
	/* border color */
	CLIO_out8(DAC_PALETTE_ADDRESS_WRITE, 0x02);
	CLIO_out8(DAC_PIXEL_DATA, ptr->col[1].r >> SHIFT); /* red */
	CLIO_out8(DAC_PIXEL_DATA, ptr->col[1].g >> SHIFT); /* green */
	CLIO_out8(DAC_PIXEL_DATA, ptr->col[1].b >> SHIFT); /* blue */
	/* cursor color 1 */
	CLIO_out8(DAC_PALETTE_ADDRESS_WRITE, 0x0F);
	CLIO_out8(DAC_PIXEL_DATA, ptr->col[3].r >> SHIFT); /* red */
	CLIO_out8(DAC_PIXEL_DATA, ptr->col[3].g >> SHIFT); /* green */
	CLIO_out8(DAC_PIXEL_DATA, ptr->col[3].b >> SHIFT); /* blue */
	/* reset palette state */
	CLIO_out8(PALETTE_STATE,
		CLIO_in8(PALETTE_STATE) & ~ACCESS_CURSOR_COLORS);
}


static void dac_show_pointer(struct kgi_display *dpy, ggi_uint x, ggi_uint y)
{
	GET_PRIVATE;

	if (!(CUR_ENABLED)) {
		CLIO_out16(MISCELLANEOUS_CONTROL,
			CLIO_in16(MISCELLANEOUS_CONTROL) | CURSOR_ENABLE);
		CUR_ENABLED = 1;
	}

	if ((CUR_X != x) || (CUR_Y != y)) {

		CUR_X = x;
		CLIO_out16(CURSOR_X, x & CURSOR_X_MASK);
		CUR_Y = y;
		CLIO_out16(CURSOR_Y, y & CURSOR_Y_MASK);
	}
}

static void dac_hide_pointer(struct kgi_display *dpy)
{
	GET_PRIVATE;

	if (CUR_ENABLED) {
		CLIO_out16(MISCELLANEOUS_CONTROL,
			CLIO_in16(MISCELLANEOUS_CONTROL) & ~CURSOR_ENABLE);
		CUR_ENABLED = 0;
	}

	CUR_X = 0x0FFF;
	CUR_Y = 0x0FFF;
}

/* We have a 8 bits CLUT (therefore, we use it !)
** (We also have a 6 bits CLUT, but well, why use it ?)
*/

static void dac_set_clut(struct kgi_display *dpy, ggi_clut *clut,
			ggi_uint first, ggi_uint cnt)
{
	ushort *data = (ushort *) (clut->data + first);
	ASSERT(cnt);

	DEBUG2("Setting %d entries of a new clut", cnt);

	/* First entry to set adress */
	CLIO_out8(DAC_PALETTE_ADDRESS_WRITE, first);
	/* Now setting all entries */
	while (cnt--) {
	  CLIO_out8(DAC_PIXEL_DATA, (*(data++)) >> SHIFT); /* red */
	  CLIO_out8(DAC_PIXEL_DATA, (*(data++)) >> SHIFT); /* green */
	  CLIO_out8(DAC_PIXEL_DATA, (*(data++)) >> SHIFT); /* blue */
	  /* auto-increment of palette adress done by the chipset */
	}
}

static void dac_get_clut(struct kgi_display *dpy,
			 ggi_clut *clut, ggi_uint first, int cnt)
{
	ushort *data = (ushort *) (clut->data + first);

	DEBUG2("Reading %d entries of the clut from %d", cnt, first);

	/* First entry to read adress */
	CLIO_out8(DAC_PALETTE_ADDRESS_READ, first);
	/* Now reading all entries */
	while (cnt--) {
	  if (!(cnt % 8)) {
		DEBUG2("Reading %i th entry of clut",cnt);
	  }
	  *(data++) = CLIO_in8(DAC_PIXEL_DATA) * MAGIC; /* red */
	  *(data++) = CLIO_in8(DAC_PIXEL_DATA) * MAGIC; /* green */
	  *(data++) = CLIO_in8(DAC_PIXEL_DATA) * MAGIC; /* blue */
	  /* auto-increment of palette adress done by the chipset */
	}
}


/*
**	GGI interface
*/
int kgim_ramdac_init(struct kgi_card *card, struct kgi_display *dpy)
{
	struct dac_private *private;

	NOTICE(DRIVER_NAME " rev " DRIVER_REV);

	if (!(private = kgi_malloc_small(sizeof(*private), GFP_KERNEL))) {

		NOTICE("Failed to allocate ramdac private data!");
		return -ENOMEM;
	}
	KGIM_PRIVATE(dpy,ramdac) = private;
	memset(private, 0, sizeof(*private)); /* Mandatory (for cur_enabled)*/

	card->ramdac=&ramdac;

	return EOK;
}

void kgim_ramdac_done(struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " removed.");
}

int
kgim_ramdac_check_mode (struct kgi_display *dpy,
                        struct kgi_mode *mode, enum kgi_tm_cmd cmd)
{
  int newclock = mode->dclk;
  
	DEBUG2("ramdac_check_mode: clock is initially %d", newclock);

  if (!((1<<mode->request.graphtype)&ramdac.modes))
   {
  	DEBUG2("couldn't handle graphtype %i",mode->request.graphtype);
 	return -E(RAMDAC,MODE_CHECK);
   }

  switch (cmd)
   {
     case CMD_PROPOSE:
	mode->lclk.mul = mode->clk.mul = 1;
        mode->lclk.div = mode->clk.div = 1;
	return EOK;

     case CMD_LOWER:
       NOTICE2("Ramdac CMD_LOWER");
	if (   (newclock < ramdac.clock[mode->request.graphtype].min)
	    || (newclock < ramdac.normal.min))
	  {
	    DEBUG2("DCLK too low (%i Hz).", newclock);
	    return -E(RAMDAC,CLK_LOW_LIMIT);
	  }

	if (ramdac.clock[mode->request.graphtype].max < newclock)
    	  newclock = ramdac.clock[mode->request.graphtype].max;
	DEBUG2("Ramdac newclock=%i",newclock);
	mode->dclk   = newclock;
	return EOK;

     case CMD_RAISE:
	if (   (newclock > ramdac.clock[mode->request.graphtype].max)
            || (newclock > ramdac.normal.max))
	  {
	    DEBUG2("DCLK too high (%i Hz).", newclock);
	    return -E(RAMDAC,CLK_HIGH_LIMIT);
	  }

	if (newclock < ramdac.clock[mode->request.graphtype].min)
	  newclock = ramdac.clock[mode->request.graphtype].min;

	if (newclock < ramdac.normal.min)
	  newclock = ramdac.normal.min;

	mode->dclk = newclock;
	return EOK;

     case CMD_CHECK:
        if ((mode->lclk.mul != 1) || (mode->lclk.div != 1)
	    /* || (mode->clk.mul  != 1) */ || (mode->clk.div  != 1))
	  return -E(RAMDAC,MODE_CHECK);

	if (   (mode->dclk < ramdac.normal.min)
	    || (mode->dclk < ramdac.clock[mode->request.graphtype].min))
	  return -E(RAMDAC,CLK_LOW_LIMIT);

	if (   (mode->dclk > ramdac.normal.max)
	    || (mode->dclk > ramdac.clock[mode->request.graphtype].max))
	  return -E(RAMDAC,CLK_HIGH_LIMIT);

	return EOK;

    default:
	INTERNALERROR;
	return -E(RAMDAC,BOGUS);
  }
}

void kgim_ramdac_set_mode(struct kgi_display *dpy,struct kgi_mode *mode)
{
	DEBUG2("ramdac_set_mode: for graphtype=%.2x",mode->request.graphtype);

	/* Disable mask on the CLUT */
	CLIO_out8(DAC_PIXEL_MASK, 0xFF);

	switch (mode->request.graphtype) {
		case GT_1BIT:
		case GT_TEXT16:
		case GT_4BIT:
		case GT_8BIT:
		case GT_15BIT:
		case GT_16BIT:
		case GT_24BIT:
		case GT_32BIT:
			break;
		default:
			INTERNALERROR;
	}

        dpy->set_clut = dac_set_clut;
	#warning pointer does not work in text mode. do emulation -- ortalo
	dpy->set_pointer_shape  = dac_set_pointer_shape;
        dpy->show_pointer = dac_show_pointer;
        dpy->hide_pointer = dac_hide_pointer;
	dpy->undo_pointer = NULL;
}

int kgim_ramdac_command(struct kgi_graphic *graph, ggi_uint cmd, ggi_uint arg)
{
	/*	this may set 
	**	-  dac_params
	**	-  hardware cursor (to be done)
	**	-  CLUT (or gamma correction table).
	*/
	ggi_clut clut;

	switch(cmd) { 
		case RAMDAC_GETINFO:
			graph->io_res=&ramdac;
			return EOK;
		case RAMDAC_SETINFO:
			if (suser()) {
				memcpy(&ramdac,graph->io_buf,sizeof(ramdac));
				return EOK;
			} else  return -EPERM;
		case RAMDAC_SETCLUT:
			{  clut.size=256;
			   clut.data=(ggi_color *)(graph->io_buf);
			   dac_set_clut(graph->dev.dpy,&clut,0,256);
			} 
			return EOK;
		case RAMDAC_GETCLUT:
			{  clut.size=256;
			   clut.data=(ggi_color *)(graph->io_buf);
			   graph->io_res=graph->io_buf;
			   dac_get_clut(graph->dev.dpy,&clut,0,256);
			} 
			return EOK;
	}
  	return -E(RAMDAC,NOSUP_ALWAYS_CANT);  /* This driver can't do this. */
}






