/*\
|*| ----------------------------------------------------------------------------
|*|	KGI kernel module internal interface definitions
|*| ----------------------------------------------------------------------------
|*|
|*|	Copyright (C)	by Steffen Seeger   [seeger@physik.tu-chemnitz.de]
|*|	1995-1997	and Andreas Beck  [becka@hp.rz.uni-duesseldorf.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ----------------------------------------------------------------------------
|*|	MAINTAINER	"Steffen Seeger, [seeger@physik.tu-chemnitz.de]"
\*/

#ifndef _GGI_MODULE_H
#define _GGI_MODULE_H

#define GRAPHDEV_MAJOR        60
#define GRAPHDEV_NAME         "graphic"

#include <ggi/system.h>
#include <ggi/kgi.h>
#include <ggi/ggi_commands.h>

extern struct kgi_display my_display;


/* !!!	These make up the new internal module interface.
** !!!	Macros are provided for 'backward compatibility', but 
** !!!	they will go away.
*/
extern void DAC_out8(struct kgi_display *dpy, uint8 val, uint8 reg);
extern uint8 DAC_in8(struct kgi_display *dpy, uint8 reg);
extern void DAC_outs8(struct kgi_display *dpy, uint8 reg, void *buf, ggi_uint cnt);
extern void DAC_ins8(struct kgi_display *dpy, uint8 reg, void *buf, ggi_uint cnt);

extern void CLK_out8(struct kgi_display *dpy, uint8 val, uint8 reg);
extern uint8 CLK_in8(struct kgi_display *dpy, uint8 reg);

#define	outbDAC(val, reg)	DAC_out8(&my_display, val, reg)
#define	inbDAC(reg)		DAC_in8(&my_display, reg)
#define	outsbDAC(reg,buf,cnt)	DAC_outs8(&my_display, reg, buf, cnt)
#define	insbDAC(reg,buf,cnt)	DAC_ins8(&my_display, reg, buf, cnt)

#define	setCLK(val)		CLK_out8(&my_display, val, 0)
	/* drive CS[0:3] lines to value	*/
extern inline int getCLK(void)	/* read state of CS[0:3] lines	*/
{
	return CLK_in8(&my_display, 0);
}

/*
extern ggi_uint outsbDDC(void *buf, ggi_uint count);
extern ggi_uint insbDDC (void *buf, ggi_uint count);
*/

#define	COMMAND_ARGS struct kgi_graphic*, ggi_uint cmd, ggi_uint arg
#define	CHECK_ARGS struct kgi_display *, struct kgi_mode *, enum kgi_tm_cmd
#define SET_ARGS   struct kgi_display *, struct kgi_mode *

extern int	kgim_monitor_init (struct kgi_card *, struct kgi_display *);
extern void	kgim_monitor_done (struct kgi_display *dpy);
extern void	kgim_monitor_reset (void);
extern int	kgim_monitor_check_mode(CHECK_ARGS *);
extern void	kgim_monitor_set_mode (SET_ARGS);
extern int	kgim_monitor_command (COMMAND_ARGS);

#define	DECLARE_SUBSYSTEM(S)	extern int kgim_##S##_init (struct kgi_card *, struct kgi_display *);\
				extern void kgim_##S##_done (struct kgi_display *dpy);\
				extern void kgim_##S##_reset (void);\
				extern int  kgim_##S##_check_mode (CHECK_ARGS);\
				extern void kgim_##S##_set_mode (SET_ARGS);\
				extern int  kgim_##S##_command (COMMAND_ARGS); 

DECLARE_SUBSYSTEM(chipset)
DECLARE_SUBSYSTEM(ramdac)
DECLARE_SUBSYSTEM(clock)
DECLARE_SUBSYSTEM(graphics);

#undef	DECLARE_SUBSYSTEM
#undef	COMMAND_ARGS
#undef	CHECK_ARGS
#undef	SET_ARGS

struct kgim_private
{
	void	*driver;
	void	*monitor;
	void	*chipset;
	void	*ramdac;
	void	*clock;
	void	*graphics;
};

#define	KGIM_PRIVATE(dpy, subsystem) \
	(((struct kgim_private *) dpy->driver_data)->subsystem)

struct kgim_options_pci 
{
	pcicfg_vaddr	dev;
	mem_addr	base0;
	mem_addr	base1;
	mem_addr	base2;
	mem_addr	base3;
	mem_addr	base4;
};

struct kgim_options_misc
{
	mem_addr	law_base;
	ggi_uint	display;
};

struct kgim_options_chipset
{
	const char	*chipset;
};

struct kgim_options
{
	struct kgim_options_misc	*misc;
	struct kgim_options_pci		*pci;
	struct kgim_options_chipset	*chipset;
};

#define	KGIM_OPTIONS(dpy, opt)	\
	(((struct kgim_options *) KGIM_PRIVATE(dpy, driver))->opt)

#endif	/* #ifdef _GGI_MODULE_H */
