/* -----------------------------------------------------------------------------
**	XTERM console parser state
** -----------------------------------------------------------------------------
**
**	Copyright (C) 1996-2000, Steffen Seeger, [seeger@physik.tu-chemnitz.de]
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** -----------------------------------------------------------------------------
**	MAINTAINER	"Steffen Seeger, [seeger@physik.tu-chemnitz.de]"
*/

#ifndef	_GGI_XTERM_H
#define	_GGI_XTERM_H

#define	MAX_COLUMNS	160
#define	TAB_STOPS	MAX_COLUMNS/(8*sizeof(ulong))


enum xterm_modes
{
	display_ctrl = scroll_last_mode,/* display control characters	*/
	toggle_meta,			/* set meta bit 0x80		*/
	utf8,				/* do utf8 decoding		*/
	applic_key,			/* keypad in application mode	*/

	cursor_key,			/* DECCKM - cursor key mode	*/
	column132,			/* 132 column mode		*/
	x10_mouse,			/* x10 compatible mouse report	*/
	textronixmode,			/* enter textronix mode		*/

	allow_132mode,			/* allow switch to 132 columns	*/
	curses_fix,			/* allow hack for curses	*/
	margin_bell,			/* margin bell			*/
	logging,			/* do logging			*/

	report_mouse,			/* report mouse			*/
	track_mouse,			/* report & hilite track mouse	*/
	ansi,				/* DECANM - ANSI mode		*/
	newline,			/* ENTER sends CR/CRLF		*/

	insert,				/* ANSI insert mode		*/
	meta_esc,			/* META chars with ESC-prefix	*/

	xterm_last_mode
};

enum xterm_flags 
{
	question = scroll_last_flag,	/* ESC?[ sequence?	*/
	num_lock,			/* num_lock key active	*/
	caps_lock,			/* CAPS-shift, locked	*/
	caps_sticky,			/* CAPS-shift, sticky	*/

	xterm_last_flag
};


enum xterm_states
{
	ESnormal,	ESesc,		ESsquare,	ESgetpars,
	ESgotpars,	ESfunckey, 	EShash,		ESset,
	ESpercent,	ESignore
};

struct xterm_state
{
	uchar		state;
	uchar		charset;	
	uchar		s_charset;
	uchar		utf_count;

	ggi_uint		npar;
	ulong		par[MAX_NPAR];

	ulong		tab_stop[TAB_STOPS];	/* tabstops		*/

	isochar		utf_char;
	unicode		*translate;

	int		s_x, s_y;	/* saved cursor position	*/
	int		p_x, p_y;	/* saved pointer position	*/

	uchar		s_g[4];
	uchar		g[4];
	uchar		colors[8];
	uchar		s_colors[8];

	ggi_attr	s_attrfl;
	ulong		s_mode;	/* saved modes			*/
};


#ifdef NEED_CONSOLE_DEFS

#define	LAT1_MAP	0
#define	GRAF_MAP	1
#define	IBMPC_MAP	2
#define	USER_MAP	3

#define	XTERM		(cons->term.xterm)

#define	COLORS		(XTERM.colors)
#define	STATE		(XTERM.state)
#define	CHARSET		(XTERM.charset)
#define	G		(XTERM.g)
#define	G0		(XTERM.g[0])
#define	G1		(XTERM.g[1])
#define	G2		(XTERM.g[2])
#define	G3		(XTERM.g[3])

#define	XNPAR		(XTERM.npar)
#define	PAR		(XTERM.par)

#define	UTF_COUNT	(XTERM.utf_count)
#define	UTF_CHAR	(XTERM.utf_char)

#define	TAB_STOP	(XTERM.tab_stop)
#define	TRANSLATE	(XTERM.translate)

#define	S_ATTRFL	(XTERM.s_attrfl)
#define	S_COLORS	(XTERM.s_colors)
#define	S_X		(XTERM.s_x)
#define	S_Y		(XTERM.s_y)
#define P_X		(XTERM.p_x)
#define P_Y		(XTERM.p_y)
#define	S_CHARSET	(XTERM.s_charset)
#define	S_G		(XTERM.s_g)
#define	S_G0		(XTERM.s_g[0])
#define	S_G1		(XTERM.s_g[1])
#define	S_G2		(XTERM.s_g[2])
#define	S_G3		(XTERM.s_g[3])

#define	S_MODE		(XTERM.s_mode)

#define	SMODE(x)	(S_MODE & (1 << x))
#define	SAVE_MODE(x)	S_MODE =   (S_MODE & ~(1 << x))\
				 | (S_MODE &  (1 << x))
#define	RESTORE_MODE(x)	S_MODE =   (S_MODE & ~(1 << x))\
				 | (S_MODE &  (1 << x))

static inline int xterm_paranoia_check(void)
{
	return (xterm_last_flag < 32) && (xterm_last_mode < 32);
}

extern void xterm_do_reset(struct kgi_console *cons, int do_reset);
extern void xterm_handle_event(struct kgi_device *dev, ggi_event *ev);
extern int xterm_do_write(struct kgi_console *cons, int from_user,
		const char *buf, int count);

#endif	/* #ifdef NEED_CONSOLE_DEFS */

#endif	/* #ifdef _GGI_XTERM_H*/
