/* 
   Graphics library for GGI.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de
   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>

#include "ggi-dl.h"
#include <ggi/ggi_commands.h>

/* 
 * We are guaranteed to have been checked already...
 */

struct ggi_visual_op2d fallback_op2d;

int GGIdlinit(ggi_visual *visual,const char *version)
{
	ggi_gc *mygc;
	
	/* map the HW GC ! */
	mygc=mmap(NULL,/*PAGE_SIZE*/4096,PROT_READ|PROT_WRITE,
			MAP_SHARED,VIS_FD(visual),MMAP_TYPE_GC);

	if (mygc==(void *)-1) mygc=NULL;
	else VIS_GC(visual)=mygc;
	DPRINT("Signature: %x\n",*(int *)VIS_GC(visual));

	/* This is only an "override library". It will try to override
	 * functions which it thinks the Accelerator can do better.
	 */

	/* Color mapping 
	 */

	/**** 2D ops ****/

	/* Save away the old operations. We might need to use them. */
	fallback_op2d=*visual->op2d;

	/* Positioning is not supported...
	 */

	/* Sprites aren't supported...
	 */

	/* Generic drawing
	 */

	if (mygc) /* These functions need the Gc and thus only work,
		     when mapping it worked. */
	{
		visual->op2d->drawline=GGIdrawline;
		visual->op2d->drawbox=GGIdrawbox;
		visual->op2d->fillscreen=GGIfillscreen;
	}
	
	visual->op2d->copybox=GGIcopybox;

	return GGI_DL_OP2D;
}

int GGIdlcleanup(ggi_visual *visual)
{
	/* FIXME unmap GC. */
	munmap(VIS_GC(visual), 4096);
	VIS_GC(visual) = NULL;
                
	return EOK;
}
		

