/* 
   Graphics library for GGI. Pixels.

   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ggi-dl.h"
#include "color-magic.h"

/************************************************************
 * Let's pretend we have a 8:8:8 (TrueColor) color cube !   *
 ************************************************************/

/* Return `8:8:8' mapping
 */
ggi_uint GGImapcolor(ggi_visual *vis,ggi_color col)
{
	return	((col.r>>(GGI_COLOR_PRECISION-8))<<16)|
		((col.g>>(GGI_COLOR_PRECISION-8))<<8)|
		 (col.b>>(GGI_COLOR_PRECISION-8));
}

/* Map `8:8:8' pixel to color
 */
int GGIunmappixel(ggi_visual *vis,ggi_uint pixel,ggi_color *col)
{
	if (col==NULL) 
		return -1;

	col->b=( pixel     & 0xFF) * MAGIC_8; 
	col->g=((pixel>> 8)& 0xFF) * MAGIC_8;
	col->r=((pixel>>16)& 0xFF) * MAGIC_8;

	return EOK;	
}

/* Pack the colors into an array
 */
int GGIpackcolors(ggi_visual *vis,void *outbuf,ggi_color *cols,int len)
{
	ggi_uint *obuf=(ggi_uint *)outbuf;
	int i;

	for (i=0;i<len;i++) 
		*(obuf++)=(ggi_uint)(ggiMapColor(vis,*(cols++)));

	return 0;
}	
	
/* Unpack into the ggi_color array the values of the pixels
 */
int GGIunpackpixels(ggi_visual *vis,void *outbuf,ggi_color *cols,int len)
{
	ggi_uint *obuf=(ggi_uint *)outbuf;
	int i;

	for (i=0;i<len;i++) 
		ggiUnmapPixel(vis,*(obuf++),(cols++));

	return 0;
}	
	
  
