/* 
 *  Graphics library for GGI.
 * 
 *  Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu
 * 
 *  If you do any modifications, I would like you to send diffs to me
 *  to allow for collecting a more and more complete set of drivers and
 *  to improve existing ones.
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * ----------------------------------------------------------------------
 * 
 *	$Log: init.c,v $
 *	Revision 1.14  1997/12/18 00:19:37  becka
 *	Killed a few bugs.
 *	
 *	Revision 1.13  1997/12/15 17:28:42  becka
 *	Various slight fixes.
 *	
 *	Revision 1.12  1997/12/14 21:56:34  becka
 *	uint/sint namespace fixed ... please test ...
 *	
 *	Revision 1.11  1997/12/13 01:28:42  becka
 *	Alpha fixes. LibGGI mode negotiation changes ...
 *	
 *	Revision 1.10  1997/12/08 17:17:46  becka
 *	Keymaps right now ...
 *	
 *	Revision 1.9  1997/12/07 16:27:55  becka
 *	LibGGI extension fixes
 *	
 *	Revision 1.8  1997/12/04 14:50:51  becka
 *	Slight fixes.
 *	
 *	Revision 1.7  1997/12/03 17:08:22  becka
 *	Fixed X ressource-keeping bug.
 *	
 *	Revision 1.6  1997/11/30 13:55:56  becka
 *	Threads fixes, Extension handling.
 *	
 *	Revision 1.5  1997/11/25 19:39:43  becka
 *	default mode made available to outside.
 *	
 *	Revision 1.4  1997/11/22 20:47:01  becka
 *	LIBGGI_DEFMODE support.
 *	
 *	Revision 1.3  1997/11/15 01:13:19  willie
 *	Changed GPL to LGPL License
 *	
 *	Revision 1.2  1997/11/04 23:11:22  becka
 *	OSF1 support. Some fixups.
 *	
 *	Revision 1.1  1997/10/12 04:53:41  tfries
 *	new subdir
 *	
 *	Revision 1.14  1997/06/14 23:48:48  ezrec
 *	Minor: Fixed some of Todd's Makefile bugs..
 *	
 *	Revision 1.13  1997/06/04 05:41:28  ezrec
 *	Minor: First pass at gamma-map support (_completely_ untested), and
 *	       libggi auto-detects the X `DISPLAY' environment variable and
 *	       uses the display-X driver. (Which doesn't work yet...)
 *
 *	Revision 1.12  1997/06/02 01:03:53  ezrec
 *	Minor: Changed the ggiOpen(NULL) default to correspond with the
 *	       the top level `make devices'
 *
 *	Revision 1.11  1997/05/27 02:23:47  ezrec
 *	Minor: Changed `LIBGGI_DEFFAULT' envvar to `LIBGGI_DISPLAY'
 *
 *	Revision 1.10  1997/05/23 23:29:56  ezrec
 *	Major: Moved event handling to the display driver, added 2.1.40 support.
 *		libggi-dynamic debugging is now environment-variable controlled.
 *
 *	Revision 1.9  1997/05/16 05:12:12  ezrec
 *	Major: Event support, and updated ggilib shim. Xggi 0.0.8 binaries should
 *		work now, with a minor caveat on mouse buttons..
 *
 *	Revision 1.8  1997/04/17 05:08:11  ezrec
 *	Major: 	New ggiOpen() semantics ie ggiOpen("display-KGI:/dev/graph3",NULL)
 *		First attempt at linear_1 frame buffer
 *		Moved thread support `to the side'
 *		Added DPRINT() support
 *		Added display-KGI driver
 *
 *	Revision 1.7  1997/03/25 04:54:11  ezrec
 *	Major: libggi now handles libggi.conf files, IBM vga driver suggests
 *	      default ioctl RAMDAC DLL, default ioctl RAMDAC DLL support,
 *	      -Wall added to the libggi-dynamic Makefiles.
 *
 *	Revision 1.6  1997/03/23 08:27:40  ezrec
 *	Major: libggi-dynamic now uses `stubs' instead of #defines so that
 *	       new versions of the library won't break old ones...
 *
 * 
 * ----------------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <fcntl.h>

#include "internal.h"

/*
 * Initalize the strutures for the library
 */
ggi_focus	_ggiFocus;
static int 	_ggiLibIsUp=0;

static char  ggiconfstub[512]=GGICONFFILE;
static char *ggiconffile=ggiconfstub+4;

int ggiDebugState=0;

int _ggiExtensionSize=0;
ggi_extension *_ggiExtension;

int ggiInit(void)
{
	int err;
	const char *str;

	_ggiLibIsUp++;
	if (_ggiLibIsUp!=1) return EOK;	/* Initialize only at first call. */

	_ggiLockInit(&_ggiFocus);
	_ggiFocus.visuals=0;
	_ggiFocus.visual=NULL;
	_ggiFocus.focus=NULL;

	str=getenv("LIBGGI_DEBUG");
	if (str!=NULL) {
		ggiDebugState=atoi(str);
		DPRINT("Debugging=%d\n",ggiDebugState);
	}

	str=getenv("LIBGGI_DEFMODE");
	if (str!=NULL) {
		int visx,visy,virx,viry,texx,texy,gmod;

		visx=ggiDefaultMode.visible.x;
		visy=ggiDefaultMode.visible.y;
		virx=ggiDefaultMode.virt.x;
		viry=ggiDefaultMode.virt.y;
		texx=ggiDefaultMode.text.x;
		texy=ggiDefaultMode.text.y;
		
		gmod=8; 
		sscanf(str,"%dx%dx%d %dx%d %dx%d",
			&visx,&visy,&gmod, &virx,&viry, &texx,&texy);

		ggiDefaultMode.visible.x=visx;
		ggiDefaultMode.visible.y=visy;
		ggiDefaultMode.virt.x   =virx;
		ggiDefaultMode.virt.y   =viry;
		ggiDefaultMode.text.x   =texx;
		ggiDefaultMode.text.y   =texy;

		switch (gmod) {
			case 1:	ggiDefaultMode.graphtype=GT_1BIT;break;
			case 4:	ggiDefaultMode.graphtype=GT_4BIT;break;
			case 8:	ggiDefaultMode.graphtype=GT_8BIT;break;
			case 15: ggiDefaultMode.graphtype=GT_15BIT;break;
			case 16: ggiDefaultMode.graphtype=GT_16BIT;break;
			case 24: ggiDefaultMode.graphtype=GT_24BIT;break;
			case 32: ggiDefaultMode.graphtype=GT_32BIT;break;
			default: fprintf(stderr,"init.c: Invalid default depth!\n");
			 break;
		}
	}

	err=_ggiLoadConfig(ggiconffile);
	if (err==0)
		return EOK;

	fprintf(stderr,"libggi: couldn't open %s - trying cwd\n", ggiconffile );
	
	err=_ggiLoadConfig("libggi.conf");
	if (err==0)
		return EOK;

	fprintf(stderr,"libggi: couldn't find ./libggi.conf\n");

	_ggiLockDestroy(&_ggiFocus);
	_ggiLibIsUp--;
	return err;
}

void ggiExit(void)
{
	ggi_extension *tmp,*next;

	if (!_ggiLibIsUp)
		return;

	if (_ggiLibIsUp!=1) return;	/* Exit only at last call */

	DPRINT("ggiExit: really destroying.\n");
	while (_ggiFocus.visual!=NULL) 
		ggiClose(_ggiFocus.visual);

	_ggiLockDestroy(&_ggiFocus);
	
	for (tmp=_ggiExtension;tmp;tmp=next)
	{
		next=tmp->next;
		free(tmp);
	}
	
	_ggiLibIsUp--;
	return;
}

void ggiPanic(const char *format,...)
{
	fprintf(stderr,"%s",format);
	ggiExit();
	exit(1);
}


/* Opens a virtual terminal as
 * a visual. Does NOT set the focus to it!
 * Use ggiSetFocus to set the focus. 
 */
ggi_visual *ggiOpen(const char *driver,...)
{
	va_list drivers;
	ggi_visual *vis;
	char *buff,*cp;
	char str[1024];

	if (!_ggiLibIsUp)
		return NULL;

	if (driver==NULL) {
		/* LIBGGI_DISPLAY _always_ overrides */
		cp=getenv("LIBGGI_DISPLAY");
		if (cp!=NULL)
			return ggiOpen(cp,NULL);

		/* Try the X display.. */
		cp=getenv("DISPLAY");
		if (cp!=NULL) {
			strcpy(str,"display-X:");
			strcat(str,cp);
			return ggiOpen(str,NULL);
		}

		/* Try the console.. */
		return ggiOpen("display-KGI:/dev/graphic",NULL);
	}

	_ggiLock(&_ggiFocus);

	vis=_ggiNewVisual();

	va_start(drivers,driver);

	for (;driver!=NULL;driver=va_arg(drivers,const char *)) {
	    DPRINT("Loading driver %s\n",driver);
		buff=strdup(driver);
		cp=strchr(buff,':');
		if (cp!=NULL)
			*(cp++)=0;
		_ggiOpenDL(vis,buff,cp);
	}

	vis->next=_ggiFocus.visual;
	_ggiFocus.visual=vis;
	_ggiUnlock(&_ggiFocus);

	va_end(drivers);

	return vis;
}
	
/* ggiClose
 *	Attempt to close the visual associated with a vt
 *      returns 0 on success, nonzero on error
 *      if focus is on the closed visual, focus is set to NULL
 */
int ggiClose(ggi_visual *visual)
{
	ggi_visual *vis,*pvis=NULL;

	if (!_ggiLibIsUp)
		return -1;

	_ggiLock(&_ggiFocus);
	for (vis=_ggiFocus.visual;vis!=NULL;pvis=vis,vis=vis->next)
		if (vis==visual)
			break;

	if (vis==NULL) {
		_ggiUnlock(&_ggiFocus);
		return -1;
	}

	if (pvis==NULL)
		_ggiFocus.visual=vis->next;
	else
		pvis->next=vis->next;

	if (_ggiFocus.focus==vis)
		_ggiFocus.focus=NULL;

	_ggiUnlock(&_ggiFocus);

	vis->next=NULL;
	_ggiDestroyVisual(vis);

	return 0;
}

/* _ggiFocus
 *	Basically used for old libggi compatablity
 */
int ggiSetFocus(ggi_visual *visual)
{
	ggi_visual *vis;

	if (!_ggiLibIsUp)
		return -1;

	_ggiLock(&_ggiFocus);
	for (vis=_ggiFocus.visual;vis!=NULL;vis=vis->next)
		if (vis==visual)
			break;

	if (vis==NULL)
		return -1;

	_ggiFocus.focus=vis;

	_ggiUnlock(&_ggiFocus);

	return 0;
}

ggi_visual *ggiGetFocus()
{
	return _ggiFocus.focus;
}

int ggiRegisterExtension(int size,
			int (*load)(ggi_visual_t,const char *,const char *),
			int (*release)(ggi_visual_t,const char *))
{
	int off;
	ggi_extension *tmp,*ext=malloc(sizeof(ggi_extension));

	off=sizeof(ggi_visual)+_ggiExtensionSize;

	_ggiExtensionSize+=size;
	
	ext->size=size;
	ext->load=load;
	ext->release=release;
	ext->next=NULL;
	
	if (_ggiExtension) {
		for (tmp=_ggiExtension;tmp->next;tmp=tmp->next) ;
		tmp->next=ext;
		ext->prev=tmp;			
	} else {
		_ggiExtension=ext;
		ext->prev=NULL;
	}

	return off;
}

