/* 
   Graphics library for GGI.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de
   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ggi-dl.h"

/***********************
 * Drawline, Bresenham *
 ***********************/

/* Avoid bankswitching. Could be done a little better ... FIXME 
 */
int GGIdrawline(ggi_visual *vis,int x1,int y1,int x2,int y2)
{ 
	int dx,dy,sx,sy,st,bla,modulus;
	int save_x1,save_y1,save_st;
	
	sx=1;dx=x2-x1;if (dx<0) {sx=-1;dx=-dx;}
	sy=1;dy=y2-y1;if (dy<0) {sy=-1;dy=-dy;}
	st=(dx-dy);bla=(dx>dy);
	if (bla) st-=dy;else st+=dx;
	if (bla&&sy>0) st--;
	if (!bla&&sx<0) st--;
	
	if (dx==0) 
		return (sy>0) ? ggiDrawVLine(vis,x1,y1,dy+1)
			      : ggiDrawVLine(vis,x2,y2,dy+1);
	if (dy==0) 
		return (sx>0) ? ggiDrawHLine(vis,x1,y1,dx+1)
			      : ggiDrawHLine(vis,x2,y2,dx+1);

	dx<<=1;dy<<=1;ggiDrawPixel(vis,x1,y1);
	save_x1=x1;save_y1=y1;save_st=st;

	for(modulus=0;modulus<4;modulus++) {

		x1=save_x1;y1=save_y1;st=save_st;

		while(x1!=x2||y1!=y2) {

			if (st<0) { 
				y1+=sy;st+=dx;
				if (!bla && (x1&3)==modulus ) 
					ggiDrawPixel(vis,x1,y1);
			} else {
				x1+=sx;st-=dy;
				if ( bla && (x1&3)==modulus )
					ggiDrawPixel(vis,x1,y1);
		 	}
		}
	}
	return EOK;
}
